"""
An example showing a graph display between cities positioned on the
Earth surface.

This graph displays the longest flight routes operated by Boing
777. The two main interests of this example are that it shows how to
build a graph of arbitrary connectivity, and that it shows how to
position data on the surface of the Earth.

The graph is created by first building a scalar scatter dataset with the
mlab.points3d command, and adding line information to it. One of the
difficulties is that the lines are specified using the indexing number of
the points, so we must 'massage' our data when loading it. A similar
technique to plot the graph is done in the :ref:`example_protein`.
Another example of graph plotting, showing a different technique to plot
the graph, can be seen on :ref:`example_delaunay_graph`.

To simplify things we do not plot the connection on the surface of the
Earth, but as straight lines going through the Earth. As a result
must use transparency to show the connection.

Data source: http://www.777fleetpage.com/777fleetpage3.htm
"""

###############################################################################
# The data. This could be loaded from a file, or scraped from a website

routes_data = """
Bombay,Atlanta
Johannesburg,Atlanta
Dubai,Los Angeles
Dubai,Houston
Dubai,San Francisco
New York,Hong Kong
Newark,Hong Kong
Doha,Houston
Toronto,Hong Kong
Bombay,Newark
Bombay,New York
Vancouver,Hong Kong
Dubai,Sao Paulo
Los Angeles,Sydney
Chicago,Delhi
"""

cities_data = """
Toronto,-79.38,43.65
Chicago,-87.68,41.84
Houston,-95.39,29.77
New York,-73.94,40.67
Vancouver,-123.13,49.28
Los Angeles,-118.41,34.11
San Francisco,-122.45,37.77
Atlanta,-84.42,33.76
Dubai,55.33,25.27
Sydney,151.21,-33.87
Hong Kong,114.19,22.38
Bombay,72.82,18.96
Delhi,77.21,28.67
Newark,-82.43,40.04
Johannesburg,28.04,-26.19
Doha,51.53,25.29
Sao Paulo,-46.63,-23.53
"""

###############################################################################
# Load the data, and put it in data structures we can use
import csv
routes_table = [i for i in csv.reader(routes_data.split('\n')[1:-1])]

# Build a dictionnary returning GPS coordinates for each city
cities_coord = dict()
for line in list(csv.reader(cities_data.split('\n')))[1:-1]:
    name, long_, lat = line
    cities_coord[name] = (float(long_), float(lat))

# Store all the coordinates of connected cities in a list also keep
# track of which city corresponds to a given index in the list. The
# connectivity information is specified as connecting the i-th point
# with the j-th.
cities = dict()
coords = list()
connections = list()
for city1, city2 in routes_table[1:-1]:
    if not city1 in cities:
        cities[city1] = len(coords)
        coords.append(cities_coord[city1])
    if not city2 in cities:
        cities[city2] = len(coords)
        coords.append(cities_coord[city2])
    connections.append((cities[city1], cities[city2]))


###############################################################################
from mayavi import mlab
mlab.figure(1, bgcolor=(0.48, 0.48, 0.48), fgcolor=(0, 0, 0),
               size=(400, 400))
mlab.clf()

###############################################################################
# Display points at city positions
import numpy as np
coords = np.array(coords)
# First we have to convert latitude/longitude information to 3D
# positioning.
lat, long = coords.T * np.pi / 180
x = np.cos(long) * np.cos(lat)
y = np.cos(long) * np.sin(lat)
z = np.sin(long)

points = mlab.points3d(x, y, z,
                     scale_mode='none',
                     scale_factor=0.03,
                     color=(0, 0, 1))

###############################################################################
# Display connections between cities
connections = np.array(connections)
# We add lines between the points that we have previously created by
# directly modifying the VTK dataset.
points.mlab_source.dataset.lines = connections
points.mlab_source.reset()
# To represent the lines, we use the surface module. Using a wireframe
# representation allows to control the line-width.
mlab.pipeline.surface(points, color=(1, 1, 1),
                              representation='wireframe',
                              line_width=4,
                              name='Connections')

###############################################################################
# Display city names
for city, index in cities.items():
    label = mlab.text(x[index], y[index], city, z=z[index],
                      width=0.016 * len(city), name=city)
    label.property.shadow = True

###############################################################################
# Display continents outline, using the VTK Builtin surface 'Earth'
from mayavi.sources.builtin_surface import BuiltinSurface
continents_src = BuiltinSurface(source='earth', name='Continents')
# The on_ratio of the Earth source controls the level of detail of the
# continents outline.
continents_src.data_source.on_ratio = 2
continents = mlab.pipeline.surface(continents_src, color=(0, 0, 0))

###############################################################################
# Display a semi-transparent sphere, for the surface of the Earth

# We use a sphere Glyph, through the points3d mlab function, rather than
# building the mesh ourselves, because it gives a better transparent
# rendering.
sphere = mlab.points3d(0, 0, 0, scale_mode='none',
                                scale_factor=2,
                                color=(0.67, 0.77, 0.93),
                                resolution=50,
                                opacity=0.7,
                                name='Earth')

# These parameters, as well as the color, where tweaked through the GUI,
# with the record mode to produce lines of code usable in a script.
sphere.actor.property.specular = 0.45
sphere.actor.property.specular_power = 5
# Backface culling is necessary for more a beautiful transparent
# rendering.
sphere.actor.property.backface_culling = True

###############################################################################
# Plot the equator and the tropiques
theta = np.linspace(0, 2 * np.pi, 100)
for angle in (- np.pi / 6, 0, np.pi / 6):
    x = np.cos(theta) * np.cos(angle)
    y = np.sin(theta) * np.cos(angle)
    z = np.ones_like(theta) * np.sin(angle)

    mlab.plot3d(x, y, z, color=(1, 1, 1),
                        opacity=0.2, tube_radius=None)

mlab.view(63.4, 73.8, 4, [-0.05, 0, 0])
mlab.show()



import os
import shutil
import sys
import unittest
import tempfile

import numpy

from mayavi import mlab
from mayavi.core.engine import Engine
from mayavi.core.off_screen_engine import OffScreenEngine
from mayavi.tools.figure import savefig, screenshot

from common import TestCase


def create_quiver3d():
    x, y, z = numpy.mgrid[1:10, 1:10, 1:10]
    u, v, w = numpy.mgrid[1:10, 1:10, 1:10]
    s = numpy.sqrt(u**2 + v**2)
    mlab.quiver3d(x, y, z, u, v, w, scalars=s)


# Note: the figure(window) size is delibrately set to be smaller than
# the required size in `savefig`, this forces the re-rendering to
# occur and catch any potential ill rendering

class TestMlabSavefigUnitTest(unittest.TestCase):

    def setUp(self):
        # Make a temporary directory for saved figures
        self.temp_dir = tempfile.mkdtemp()
        self.filename = os.path.join(self.temp_dir, "saved_figure.png")

        # this ensures that the temporary directory is removed
        self.addCleanup(self.remove_tempdir)

    def remove_tempdir(self):
        shutil.rmtree(self.temp_dir)

    def setup_engine_and_figure(self, engine):
        # Set up a Engine/OffScreenEngine/... for the test case
        self.engine = engine

        if not engine.running:
            engine.start()

        # figure size is set to be small to force re-rendering
        engine.new_scene(size=(90, 100))
        self.figure = engine.current_scene

        # the clean up function will close all figures and stop the engine
        self.addCleanup(self.cleanup_engine, self.engine)

    def realize(self, figure):
        """ Put the figure's window on screen so VTK gives it a size.

        A scene that has never been shown leaves vtkRenderWindow at 0x0: the Qt
        widget reports the size it was asked for, but VTK only learns one from a
        resize event, and rendering does not supply it.  mlab.screenshot reads
        render_window.size, so it would try to reshape the grab into (0, 0, 3).
        """
        from pyface.api import GUI

        figure.scene._vtk_control.window().show()
        for _ in range(20):
            GUI.process_events()
            if all(figure.scene.render_window.size):
                return
        raise AssertionError('the scene window never got a size')

    def cleanup_engine(self, engine):
        """ Close all scenes in the engine and stop it """
        scenes = [scene for scene in engine.scenes]
        for scene in scenes:
            engine.close_scene(scene)
        engine.stop()

    @unittest.skipIf(os.environ.get("TRAVIS", False),
                     ("Offscreen rendering is not tested on Travis "
                      "due to lack of GLX support"))
    def test_many_savefig_offscreen(self):
        """Test if savefig works with off_screen_rendering and Engine"""
        engine = Engine()
        for _ in range(5):
            self.setup_engine_and_figure(engine)

            # Use off-screen rendering
            self.figure.scene.off_screen_rendering = True

            # Set up the scene
            create_quiver3d()

            # save the figure
            savefig(self.filename, size=(131, 217),
                    figure=self.figure)

    def _get_pixel_ratio(self, fig):
        return getattr(fig.scene._vtk_control, '_pixel_ratio', 1.0)

    def test_mlab_screenshot(self):
        # Given
        engine = Engine()
        self.setup_engine_and_figure(engine)
        create_quiver3d()
        self.realize(self.figure)
        sz = self.figure.scene.get_size()
        pixel_ratio = self._get_pixel_ratio(self.figure)
        sz = (sz[0]*pixel_ratio, sz[1]*pixel_ratio)

        for aa in (False, True):
            # When
            data = screenshot(mode='rgb', antialiased=aa)
            # Then
            self.assertEqual(data.shape, (sz[1], sz[0], 3))

            # When
            data = screenshot(mode='rgba', antialiased=aa)
            # Then
            self.assertEqual(data.shape, (sz[1], sz[0], 4))


class TestMlabSavefig(TestCase):

    def test(self):
        self.main()

    def do(self):
        suite = unittest.TestLoader().loadTestsFromTestCase(
            TestMlabSavefigUnitTest)

        result = unittest.TextTestRunner().run(suite)

        if result.errors or result.failures:
            sys.exit(1)


if __name__ == "__main__":
    t = TestMlabSavefig()
    t.test()



# Author: Prabhu Ramachandran, Gael Varoquaux
# Copyright (c) Enthought, Inc.
# License: BSD Style.
""" A tool for easy and interactive visualization of data.
    Part of the Mayavi project of the Enthought Tool Suite.
"""

try:
    from importlib.metadata import version

    __version__ = version("mayavi")
except Exception:
    __version__ = "0.0.0"

# Keep Qt off the wayland platform, where VTK render windows cannot embed.
# This has to happen before pyface creates the QApplication, which it does
# as an import side effect deeper in mayavi (see tvtk/WORKAROUNDS.md).
from tvtk.qt_x11 import steer_qt_to_x11 as _steer_qt_to_x11
_steer_qt_to_x11()
del _steer_qt_to_x11


def _jupyter_nbextension_paths():
    return [dict(
        section="notebook",
        # the path is relative to the `mayavi` directory
        src="tools/static",
        # directory in the `nbextension/` namespace
        dest="mayavi",
        require="mayavi/x3d/x3dom"
    )]



#!/usr/bin/env python
"""
An example of how to modify the data visualized  via an interactive dialog.

A dialog is created via `TraitsUI
<https://docs.enthought.com/traitsui/>`_ from an object (MyModel).
Some attributes of the objects are represented on the dialog: first a
Mayavi scene, that will host our visualization, and two parameters that
control the data plotted.

A curve is plotted in the embedded scene using the associated
mlab.points3d function. The visualization object created is stored
as an attribute on the main MyModel object, to modify it inplace later.

When the `n_meridional` and `n_longitudinal` attributes are modified, eg via
the slide bars on the dialog, the curve is recomputed, and the
visualization is updated by modifying inplace the stored plot
object (see :ref:`mlab-animating-data`).

This example is discussed in details in the section
:ref:`embedding_mayavi_traits`.
"""
# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.


from numpy import arange, pi, cos, sin

from traits.api import HasTraits, Range, Instance, \
        on_trait_change
from traitsui.api import View, Item, Group

from mayavi.core.api import PipelineBase
from mayavi.core.ui.api import MayaviScene, SceneEditor, \
                MlabSceneModel


dphi = pi/1000.
phi = arange(0.0, 2*pi + 0.5*dphi, dphi, 'd')

def curve(n_mer, n_long):
    mu = phi*n_mer
    x = cos(mu) * (1 + cos(n_long * mu/n_mer)*0.5)
    y = sin(mu) * (1 + cos(n_long * mu/n_mer)*0.5)
    z = 0.5 * sin(n_long*mu/n_mer)
    t = sin(mu)
    return x, y, z, t


class MyModel(HasTraits):
    n_meridional    = Range(0, 30, 6, )#mode='spinner')
    n_longitudinal  = Range(0, 30, 11, )#mode='spinner')

    scene = Instance(MlabSceneModel, ())

    plot = Instance(PipelineBase)


    # When the scene is activated, or when the parameters are changed, we
    # update the plot.
    @on_trait_change('n_meridional,n_longitudinal,scene.activated')
    def update_plot(self):
        x, y, z, t = curve(self.n_meridional, self.n_longitudinal)
        if self.plot is None:
            self.plot = self.scene.mlab.plot3d(x, y, z, t,
                                tube_radius=0.025, colormap='Spectral')
        else:
            self.plot.mlab_source.trait_set(x=x, y=y, z=z, scalars=t)


    # The layout of the dialog created
    view = View(Item('scene', editor=SceneEditor(scene_class=MayaviScene),
                     height=250, width=300, show_label=False),
                Group(
                        '_', 'n_meridional', 'n_longitudinal',
                     ),
                resizable=True,
                )

my_model = MyModel()
my_model.configure_traits()



#!/usr/bin/env python
"""
This script demonstrates how one can use the Mayavi application framework
without displaying Mayavi's UI.

Note: look at the end of this file to see how the non gui plugin is
chosen instead of the default gui mayavi plugin.

This should be run as::

  $ python nongui.py

Or::

  $ mayavi2 script.py

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Standard library imports
from os.path import join, abspath

# Enthought library imports
from mayavi.scripts.util import get_data_dir
from mayavi.plugins.app import Mayavi, get_non_gui_plugins

class MyApp(Mayavi):
    def run(self):
        """This is executed once the application GUI has started.
        *Make sure all other MayaVi specific imports are made here!*
        """

        # Various imports to do different things.
        from mayavi.sources.vtk_file_reader import VTKFileReader
        from mayavi.modules.outline import Outline
        from mayavi.modules.axes import Axes
        from mayavi.modules.grid_plane import GridPlane
        from mayavi.modules.image_plane_widget import ImagePlaneWidget
        from mayavi.modules.text import Text
        from mayavi.modules.contour_grid_plane import ContourGridPlane
        from mayavi.modules.iso_surface import IsoSurface

        script = self.script

        # Create a new scene.
        script.new_scene()

        # Read a VTK (old style) data file.
        r = VTKFileReader()
        r.initialize(join(get_data_dir(abspath(__file__)),
                          'heart.vtk'))
        script.add_source(r)

        # Put up some text.
        t = Text(text='MayaVi rules!', x_position=0.2, y_position=0.9, width=0.8)
        t.property.color = 1, 1, 0  # Bright yellow, yeah!
        script.add_module(t)

        # Create an outline for the data.
        o = Outline()
        script.add_module(o)

        # Create an axes for the data.
        a = Axes()
        script.add_module(a)

        # Create three simple grid plane modules.
        # First normal to 'x' axis.
        gp = GridPlane()
        script.add_module(gp)
        # Second normal to 'y' axis.
        gp = GridPlane()
        gp.grid_plane.axis = 'y'
        script.add_module(gp)
        # Third normal to 'z' axis.
        gp = GridPlane()
        script.add_module(gp)
        gp.grid_plane.axis = 'z'

        # Create one ImagePlaneWidget.
        ipw = ImagePlaneWidget()
        script.add_module(ipw)
        # Set the position to the middle of the data.
        ipw.ipw.slice_position = 16

        # Create one ContourGridPlane normal to the 'x' axis.
        cgp = ContourGridPlane()
        script.add_module(cgp)
        # Set the position to the middle of the data.
        cgp.grid_plane.axis = 'y'
        cgp.grid_plane.position = 15

        # An isosurface module.
        iso = IsoSurface(compute_normals=True)
        script.add_module(iso)
        iso.contour.contours = [200.0]

        # Set the view.
        s = script.engine.current_scene
        cam = s.scene.camera
        cam.azimuth(45)
        cam.elevation(15)
        s.render()


if __name__ == '__main__':
    m = MyApp()
    # Get the default non GUI plugins.
    plugins = get_non_gui_plugins()
    # Start the app with these plugins.
    m.main(plugins=plugins)



"""Tests for the Text3D module

"""
import os
import shutil
import sys
import unittest
import tempfile

import numpy
from PIL import Image

from mayavi import mlab

from common import TestCase


class TestText3DUnitTest(unittest.TestCase):

    def setUp(self):
        # Make a temporary directory for saved figures
        self.temp_dir = tempfile.mkdtemp()
        self.filename = os.path.join(self.temp_dir, "saved_figure.png")

        # this ensures that the temporary directory is removed
        self.addCleanup(self.remove_tempdir)

    def remove_tempdir(self):
        shutil.rmtree(self.temp_dir)

    def mlab_close_all(self):
        mlab.close(all=True)

    def test_text3d(self):
        """Test if Text3D shows"""
        # the points3d is there to provide data for
        # attaching the text3d module.  Opacity is set to
        # zero so that the image should only show
        # the text3d and we look for the pixels
        mlab.points3d(0., 0., 0., opacity=0.)

        self.addCleanup(self.mlab_close_all)

        mlab.text3d(0., 0., 0., "X")
        mlab.savefig(self.filename)

        self.check()

    def check(self):
        image = numpy.array(Image.open(self.filename))[:, :, :3]

        # Check that the image has pixel different from the background
        # Supposedly the first pixel is the background color
        if not (numpy.sum(image != image[0, 0], axis=2) == 3).any():
            message = "Looks like the Text3D is not shown"
            self.fail(message)


class TestText3D(TestCase):

    def test(self):
        self.main()

    def do(self):
        suite = unittest.TestLoader().loadTestsFromTestCase(
            TestText3DUnitTest)

        result = unittest.TextTestRunner().run(suite)

        if result.errors or result.failures:
            sys.exit(1)


if __name__ == "__main__":
    t = TestText3D()
    t.test()



"""
Render the examples to images and adds them to the documentation.
"""

# Standard library imports
import glob
import inspect
import os
import re
from io import StringIO
from pathlib import Path
import shutil
import subprocess
import sys
import tempfile
import time
import token, tokenize
import textwrap
import itertools

import numpy as np

# Enthought imports
from mayavi import mlab

# A global counter, for subsitutions.
global_counter = itertools.count()

EXAMPLE_DIR = '../../examples/mayavi'

# Examples whose figure differs from run to run, so that re-rendering would flip
# the published image back and forth for no gain.  The first two draw
# overlapping translucent actors, which VTK composites slightly differently from
# one run to the next: tvtk_in_mayavi about one run in five, and magnetic_field
# -- which culls front faces to fake transparency -- in three of the last four
# CI runs, a third of the image at a time.  wx_mayavi_embed_in_notebook is the
# odd one out: it is a screenshot of a wx window, and the notebook it draws
# lands differently often enough to have come back changed in two of the four
# CI runs after it was added, on the committed bytes both times, on main and on
# a branch that had not touched it.  See the note in CLAUDE.md for what was
# tried on the first two.
FLAKY_EXAMPLES = frozenset({'tvtk_in_mayavi', 'magnetic_field',
                            'wx_mayavi_embed_in_notebook'})

# Examples that cannot be rendered unattended, and why.
SKIP_EXAMPLES = {
    # Its two scenes are windows of their own rather than panels of the dialog
    # the example opens, and the first of them -- the one anything looking for
    # a scene finds first -- is the empty one it creates "just for kicks", so
    # shooting it would put a blank figure in the gallery.
    'standalone': 'builds its own Engine and puts each scene in its own window',
    'user_mayavi': 'is loaded by the mayavi2 application, not run on its own',
    'zzz_reader': 'registers a reader; there is nothing to show',
}

# No example takes anything like this long; it is here so that one which will
# never finish fails the build instead of hanging it.
EXAMPLE_TIMEOUT = 60

# ...except lucy, which on a cold cache downloads 307 MB and unpacks 523 MB from
# it before it can render.  Once the data is there it is as quick as the rest.
EXAMPLE_TIMEOUTS = {'lucy': 300}

# Examples that were meant to produce a figure and did not.  A failure only
# prints and moves on, so that one broken example does not cost the gallery
# every later figure -- which also means it leaves the committed image in place
# and looks, in the diff, exactly like an example that rendered unchanged.
# render_docs.py turns a non-empty list into a failed build at the very end.
RENDER_FAILURES = []


def keep_windows_in_background():
    """ Stops the windows we open from stealing focus and swallowing keystrokes.

        Rendering has to paint on screen, but nothing says it has to happen in
        front of whatever the user is doing: WA_ShowWithoutActivating keeps the
        window from taking focus, and WindowStaysOnBottomHint keeps it behind
        the other windows.  The attribute has to be set before the window is
        shown, hence patching setVisible rather than fixing up windows
        afterwards.

        The hint, though, goes on the QWindow *after* the show, not on the
        QWidget before it.  QWidget.setWindowFlag recreates the window's native
        window, and on X11 destroying a window destroys its children -- which
        takes out the X window a scene widget was created with.  Qt goes on
        handing out the dead id (winId() returns it, and neither create() nor
        another destroy()/create() replaces it), so the next paint has VTK ask
        the server about a window that is gone; XGetWindowAttributes fails,
        VTK's CreateAWindow carries on with a null XVisualInfo, and the
        glXCreateContext failsafe at the end of it dereferences the null.  That
        is what used to take qt_embedding -- the one example that shows a
        window it has already built a scene into -- down under Xvfb and on real
        X11 alike.  Setting the hint on the QWindow leaves the native windows
        alone and keeps the same behaviour.
    """
    from pyface.qt import QtCore, QtGui

    def enum(holder, name):
        # PyQt6 scopes these, PySide6 offers them both ways
        return getattr(QtCore.Qt, name, None) or getattr(holder, name)

    # traitsui's dialogs and pyface's windows raise themselves too
    QtGui.QWidget.raise_ = lambda self: None
    QtGui.QWidget.activateWindow = lambda self: None

    no_activate = enum(QtCore.Qt.WidgetAttribute, 'WA_ShowWithoutActivating')
    at_the_back = enum(QtCore.Qt.WindowType, 'WindowStaysOnBottomHint')
    real_set_visible = QtGui.QWidget.setVisible

    def set_visible(self, visible):
        # testAttribute keeps this to once per window, so that re-showing one
        # does not push the hint down a second time
        first_show = (visible and self.isWindow()
                      and not self.testAttribute(no_activate))
        if first_show:
            self.setAttribute(no_activate, True)
        real_set_visible(self, visible)
        if first_show:
            handle = self.windowHandle()
            if handle is not None:
                handle.setFlag(at_the_back, True)

    QtGui.QWidget.setVisible = set_visible


def should_render(short_file_name, image_file):
    """ Whether to (re-)render an example, honouring MAYAVI_RENDER_FLAKY.
    """
    if short_file_name not in FLAKY_EXAMPLES:
        return True
    if not os.path.exists(image_file):
        return True  # nothing committed to fall back on
    return os.environ.get('MAYAVI_RENDER_FLAKY', '').lower() \
        not in ('', '0', 'false', 'no')


def is_wx_example(filename):
    """ Whether the example is written against wx rather than Qt.

        Read from the source rather than kept in a list, so that a wx example
        added later is picked up on its own.  It cannot go through
        _code_without_comments, which joins tokens with nothing between them.
    """
    return re.search(r'^\s*(?:import|from)\s+wx\b', Path(filename).read_text(),
                     re.MULTILINE) is not None


def is_dialog_example(filename):
    """ Whether the example puts its scene in a TraitsUI dialog.
    """
    code_only = _code_without_comments(filename)
    return 'configure_traits(' in code_only or 'edit_traits(' in code_only


def is_app_example(filename):
    """ Whether the example runs itself inside the Mayavi2 application.

        ``@mayavi2.standalone`` starts Envisage and its event loop, so such an
        example has to be shot the way a dialog is -- with the Qt loop stubbed
        out -- rather than through ``run_mlab_file``, which would block in
        ``Mayavi.main()`` until the render timed out.  Its figure is the
        workbench window, not an mlab one.
    """
    return 'mayavi2.standalone' in _code_without_comments(filename)


def _code_without_comments(filename):
    tokens = tokenize.generate_tokens(StringIO(Path(filename).read_text()).readline)
    return ''.join([tok_content
                    for tok_type, tok_content, _, _, _ in tokens
                    if not token.tok_name[tok_type] in ('COMMENT', 'STRING')])


def settle_layout(widget):
    """ Makes Qt apply the geometry now, instead of when it gets round to
        delivering the LayoutRequest events it has posted.

        Every layout between the widget and its window has to be activated,
        not just the window's; mne/viz/backends/_qt.py does the same.
    """
    seen = set()
    while widget is not None and widget not in seen:
        layout = widget.layout()
        if layout is not None:
            layout.activate()
        seen.add(widget)
        widget = widget.parentWidget()


def _scene_widgets(widget, found=None):
    """ Every VTK render widget below ``widget``.

        Matched on the attribute rather than the class name, which differs
        between the QVTKRenderWindowInteractor subclasses.
    """
    from pyface.qt import QtGui

    found = [] if found is None else found
    if hasattr(widget, '_RenderWindow'):
        found.append(widget)
    for child in widget.children():
        if isinstance(child, QtGui.QWidget):
            _scene_widgets(child, found)
    return found


def _window_holding_a_scene():
    """ The visible top-level window that holds a scene, if there is one.

        For an example that runs inside the Mayavi2 application this is the
        workbench window: it opens no dialog we could have caught, so the only
        way to find its figure is to look at what ended up on screen.
    """
    from pyface.qt import QtGui

    for widget in QtGui.QApplication.topLevelWidgets():
        if widget.isVisible() and _scene_widgets(widget):
            return widget
    return None


class NoSceneInDialog(RuntimeError):
    """ Raised for an example whose dialog holds only controls, the scene
        being in an mlab figure of its own.
    """


def capture_dialog(filename, image_file):
    """ Shoots a TraitsUI example: the dialog, with its scenes painted in.

        QWidget.grab() draws the Qt widgets but not the native surface the VTK
        scene renders onto -- on macOS that leaves a black rectangle -- so each
        scene is rendered separately and pasted over its own place in the shot.
    """
    from pyface.qt import QtCore, QtGui, QtTest
    from traits.api import Button, HasTraits
    from tvtk.api import tvtk

    created = []
    real_edit, real_configure = HasTraits.edit_traits, HasTraits.configure_traits
    old_show = mlab.show

    def edit_traits(self, *args, **kwargs):
        # non-modal, so the example returns to us -- but an example that asks
        # to be embedded ('subpanel') means it, and prising the panel out into
        # a window of its own would shoot the panel instead of the window
        kwargs.setdefault('kind', 'live')
        ui = real_edit(self, *args, **kwargs)
        created.append(ui)
        return ui

    def settle():
        settle_layout(created[-1].control if created else None)
        app = QtGui.QApplication.instance()
        for _ in range(5):
            app.processEvents()

    # a few examples run the Qt loop themselves rather than through mlab.show
    real_execs = {name: getattr(QtGui.QApplication, name)
                  for name in ('exec', 'exec_')
                  if hasattr(QtGui.QApplication, name)}
    for name in real_execs:
        setattr(QtGui.QApplication, name, lambda *a, **k: 0)

    mlab.show = lambda func=None: None
    HasTraits.edit_traits = edit_traits
    HasTraits.configure_traits = lambda self, *a, **k: edit_traits(self)
    np.random.seed(0)
    try:
        exec(compile(Path(filename).read_text(), filename, 'exec'),
             {'__name__': '__main__', '__file__': os.path.abspath(filename)})
        ui = created[-1] if created else None
        if ui is not None:
            # for an embedded panel this is the window the example built round it
            dialog = ui.control.window()
        else:
            # an example driving the Mayavi2 application opens no dialog of its
            # own: its figure is whichever window Envisage put the scene in
            dialog = _window_holding_a_scene()
            if dialog is None:
                raise RuntimeError('%s opened no dialog' % filename)
        dialog.show()
        # the scene is only built once the window is really on screen, and
        # several examples plot from a scene.activated handler
        QtTest.QTest.qWaitForWindowExposed(dialog.windowHandle() or dialog, 5000)
        settle()

        obj = ui.context.get('object') if ui is not None else None
        if obj is not None:      # a few examples only plot when clicked
            for name in obj.trait_names():
                if isinstance(obj.trait(name).trait_type, Button):
                    setattr(obj, name, True)
            settle()

        scenes = _scene_widgets(dialog)
        if not scenes:
            raise NoSceneInDialog(filename)
        shot = dialog.grab()
        painter = QtGui.QPainter(shot)
        try:
            for widget in scenes:
                with tempfile.NamedTemporaryFile(suffix='.png') as rendered:
                    rw = widget._RenderWindow
                    to_image = tvtk.WindowToImageFilter(
                        input=rw, read_front_buffer=True)
                    writer = tvtk.PNGWriter(file_name=rendered.name)
                    writer.set_input_data(to_image.output)
                    # as in TVTKScene: the front buffer is the one VTK resolves
                    # the multisamples into properly, and the swap is what
                    # fills it, so it has to be left alone
                    rw.Render()
                    to_image.update()
                    writer.write()
                    painter.drawImage(
                        QtCore.QRect(widget.mapTo(dialog, QtCore.QPoint(0, 0)),
                                     widget.size()),
                        QtGui.QImage(rendered.name))
        finally:
            painter.end()
        shot.save(image_file)
    finally:
        for ui in created:
            ui.dispose()
        HasTraits.edit_traits = real_edit
        HasTraits.configure_traits = real_configure
        mlab.show = old_show
        for name, real in real_execs.items():
            setattr(QtGui.QApplication, name, real)


def _wx_scene_controls(window, found=None):
    """ Every VTK render widget below ``window``, wx side.

        VTK's wx interactor is what tvtk builds its scene on, and GetRenderWindow
        is the part of it worth matching on.
    """
    found = [] if found is None else found
    if hasattr(window, 'GetRenderWindow'):
        found.append(window)
    for child in window.GetChildren():
        _wx_scene_controls(child, found)
    return found


def _wx_scenes_on_screen(frame, filename, timeout=EXAMPLE_TIMEOUT / 4):
    """ Waits for the frame's scenes to be on screen, and returns them.

        The Qt path waits with QTest.qWaitForWindowExposed; wx has nothing
        equivalent, so wait on the thing that is actually wanted -- a scene the
        window server has put up -- rather than on a fixed number of turns
        round the event loop.  Update() then paints what is pending now, rather
        than whenever wx would next have got to it.

        A notebook's hidden tab is left out: it sits at the same place as the
        visible one, with an unrealised render window that would otherwise be
        scaled over the top of it.
    """
    import wx

    app = wx.GetApp()
    deadline = time.monotonic() + timeout
    while True:
        app.Yield()                 # processes everything currently pending
        scenes = [scene for scene in _wx_scene_controls(frame)
                  if scene.IsShownOnScreen()]
        if scenes and frame.IsShownOnScreen():
            frame.Layout()
            # IsShownOnScreen means the widget and its parents are shown, not
            # that the server has mapped and exposed the window -- there is no
            # wx qWaitForWindowExposed.  Update() paints the pending regions
            # synchronously, so the blit that follows has something to copy.
            frame.Update()
            return scenes
        if time.monotonic() >= deadline:
            raise NoSceneInDialog(filename)


def capture_wx_dialog(filename, image_file):
    """ Shoots a wx example: the frame, with its scenes painted in.

        The Qt counterpart with the toolkit swapped out -- MainLoop for exec_,
        WindowDC onto a MemoryDC for QWidget.grab, DrawBitmap for
        QPainter.drawImage -- and for the same reason: the blit copies the wx
        widgets but not the native surface VTK draws on.
    """
    import wx
    from tvtk.api import tvtk

    frames = []
    real_frame_init = wx.Frame.__init__
    real_main_loop = wx.App.MainLoop
    old_show = mlab.show

    def frame_init(self, *args, **kwargs):
        real_frame_init(self, *args, **kwargs)
        frames.append(self)

    wx.Frame.__init__ = frame_init
    wx.App.MainLoop = lambda self: None   # these examples run the loop
    mlab.show = lambda func=None: None
    np.random.seed(0)
    try:
        exec(compile(Path(filename).read_text(), filename, 'exec'),
             {'__name__': '__main__', '__file__': os.path.abspath(filename)})
        if not frames:
            raise RuntimeError('%s opened no frame' % filename)
        frame = frames[-1]
        frame.Show()
        scenes = _wx_scenes_on_screen(frame, filename)

        width, height = frame.GetSize()
        shot = wx.Bitmap(width, height)
        canvas = wx.MemoryDC(shot)
        canvas.Blit(0, 0, width, height, wx.WindowDC(frame), 0, 0)
        try:
            for control in scenes:
                render_window = control.GetRenderWindow()
                position = frame.ScreenToClient(control.GetScreenPosition())
                size = control.GetSize()
                with tempfile.NamedTemporaryFile(suffix='.png') as rendered:
                    to_image = tvtk.WindowToImageFilter(
                        input=render_window, read_front_buffer=True)
                    writer = tvtk.PNGWriter(file_name=rendered.name)
                    writer.set_input_data(to_image.output)
                    # as in TVTKScene: the front buffer is the one VTK resolves
                    # the multisamples into properly, and the swap is what
                    # fills it, so it has to be left alone
                    render_window.Render()
                    to_image.update()
                    writer.write()
                    image = wx.Image(rendered.name)
                    if (image.GetWidth(), image.GetHeight()) != tuple(size):
                        # the render window is in pixels, the widget in points
                        image = image.Scale(size[0], size[1],
                                            wx.IMAGE_QUALITY_HIGH)
                    canvas.DrawBitmap(wx.Bitmap(image), position[0],
                                      position[1])
        finally:
            canvas.SelectObject(wx.NullBitmap)
        kind = (wx.BITMAP_TYPE_JPEG
                if image_file.lower().endswith(('.jpg', '.jpeg'))
                else wx.BITMAP_TYPE_PNG)
        shot.SaveFile(image_file, kind)
    finally:
        wx.Frame.__init__ = real_frame_init
        wx.App.MainLoop = real_main_loop
        mlab.show = old_show


def capture_one(filename, image_file):
    """ Renders one example, the way that suits it.  Runs in the child.
    """
    if is_wx_example(filename):
        # keep_windows_in_background is Qt-only, and pyface.qt need not even
        # import in a toolkit=wx child
        return capture_wx_dialog(filename, image_file)
    keep_windows_in_background()
    # an example that also draws with matplotlib would block in pyplot.show();
    # the environment only affects this child, not what users get
    os.environ.setdefault('MPLBACKEND', 'Agg')
    if is_dialog_example(filename) or is_app_example(filename):
        try:
            return capture_dialog(filename, image_file)
        except NoSceneInDialog:
            pass    # its dialog only holds controls; shoot the figure instead
    run_mlab_file(filename, image_file=image_file)


def capture_in_subprocess(filename, image_file):
    """ Renders one example in a child process, so that it can be killed.

        An example that reaches a Qt event loop never hands control back to the
        interpreter, so a signal-based timeout in this process would never be
        delivered -- only killing the process actually works.  Isolation also
        keeps one example's leftover scene state out of the next one's figure.
    """
    here = os.path.dirname(os.path.abspath(__file__))
    code = ('import mayavi, tvtk.api, sys\n'          # bind before the doc
            'sys.path.insert(0, %r)\n'                # sources can shadow them
            'from render_examples import capture_one\n'
            'capture_one(%r, %r)\n' % (here, filename, image_file))
    short_name = os.path.splitext(os.path.basename(filename))[0]
    # the toolkit is per process, which is exactly what the isolation buys us:
    # a wx example gets a wx child and the rest of the gallery stays on Qt
    env = dict(os.environ)
    if is_wx_example(filename):
        env['ETS_TOOLKIT'] = 'wx'
    subprocess.run([sys.executable, '-P', '-c', code], check=True,
                   timeout=EXAMPLE_TIMEOUTS.get(short_name, EXAMPLE_TIMEOUT),
                   capture_output=True, text=True, env=env)


def capture_example(filename, short_file_name, image_file):
    """ Renders one example's figure, picking the way that suits it.

        Anything that cannot be rendered is left with whatever image is
        committed, so the gallery keeps its thumbnail either way.
    """
    if short_file_name in SKIP_EXAMPLES:
        print("Skipping %s: %s" % (filename, SKIP_EXAMPLES[short_file_name]))
        return
    if not should_render(short_file_name, image_file):
        print("Keeping the committed image for %s; it does not render "
              "reproducibly (set MAYAVI_RENDER_FLAKY=1 to redo it)" % filename)
        return
    if not (is_dialog_example(filename) or is_mlab_example(filename)
            or is_app_example(filename)):
        print("Skipping %s: it neither shows a figure nor opens a dialog"
              % filename)
        return
    print("Generating images for %s" % filename, flush=True)
    try:
        capture_in_subprocess(filename, image_file)
        if not os.path.exists(image_file):
            raise RuntimeError('rendered without leaving an image behind')
    except Exception as exc:
        # one broken example should not cost the gallery every later figure
        RENDER_FAILURES.append('%s: %s: %s'
                               % (filename, type(exc).__name__, exc))
        print("Could not render %s: %s: %s"
              % (filename, type(exc).__name__, exc))
        # the child's traceback is the only thing that says why, and it is not
        # part of str(CalledProcessError) -- without it a CI-only failure is
        # undiagnosable from the log
        for stream in (getattr(exc, 'stderr', None), getattr(exc, 'stdout', None)):
            if stream:
                text = stream if isinstance(stream, str) else stream.decode(
                    errors='replace')
                # faulthandler's dump ends with a list of every loaded
                # extension module, thousands of characters of it, which a
                # tail would show instead of the stack that matters
                lines = [line for line in text.strip().splitlines()
                         if not line.startswith('Extension modules:')]
                # both ends: a chained exception puts the cause that actually
                # explains the failure at the top, a crash dump puts the stack
                # at the bottom, and a tail alone loses the first of those
                if len(lines) > 70:
                    lines = (lines[:30] + ['    ... %d lines omitted ...'
                                           % (len(lines) - 70)] + lines[-40:])
                print(textwrap.indent('\n'.join(lines), '    '), flush=True)


def is_mlab_example(filename):
    tokens = tokenize.generate_tokens(StringIO(Path(filename).read_text()).readline)
    code_only = ''.join([tok_content
                            for tok_type, tok_content, _, _, _  in tokens
                            if not token.tok_name[tok_type] in ('COMMENT',
                                                                'STRING')])
    return ('mlab.show()' in code_only)


def realize_scene_window(scene):
    """ Puts a scene's window on screen at the size the example asked for.

        The render window only picks the size up from a resize event, so until
        the window is really on screen it stays at VTK's 300x300 default -- and
        a figure that was never shown at all reports 0x0, which is what
        mlab.screenshot cannot reshape.
    """
    control = getattr(scene, '_vtk_control', None)
    if control is None:
        return
    from pyface.qt import QtTest
    window = control.window()
    window.show()
    QtTest.QTest.qWaitForWindowExposed(window.windowHandle() or window, 5000)
    settle_layout(control)


def run_mlab_file(filename, image_file):
    ## XXX: Monkey-patch mlab.show, so that we keep control of the
    ## the mainloop
    old_show = mlab.show
    old_screenshot = mlab.screenshot

    def my_show(func=None):
        pass

    def my_screenshot(figure=None, *args, **kwargs):
        # an example that screenshots itself does so before anything has put
        # the figure on screen, where VTK still reports it as 0x0
        realize_scene_window((figure if figure is not None else mlab.gcf()).scene)
        return old_screenshot(figure, *args, **kwargs)

    mlab.show = my_show
    mlab.screenshot = my_screenshot
    mlab.clf()
    np.random.seed(0)  # so an example plotting random data renders the same way
    e = mlab.get_engine()
    e.close_scene(mlab.gcf())
    exec(
        compile(Path(filename).read_text(), filename, 'exec'),
        {'__name__': '__main__', '__file__': os.path.abspath(filename)}
    )
    realize_scene_window(mlab.gcf().scene)
    mlab.savefig(image_file)
    size = mlab.gcf().scene.get_size()
    for scene in e.scenes:
        e.close_scene(scene)
    mlab.show = old_show
    mlab.screenshot = old_screenshot


def extract_docstring(filename):
    # Extract a module-level docstring, if any
    lines = Path(filename).read_text().splitlines(keepends=True)
    start_row = 0
    if lines[0].startswith('#!'):
        lines.pop(0)
        start_row = 1

    docstring = ''
    first_par = ''
    li = lines.__iter__()
    li_next = li.__next__ if hasattr(li, '__next__') else li.next
    tokens = tokenize.generate_tokens(li_next)
    for tok_type, tok_content, _, (erow, _), _ in tokens:
        tok_type = token.tok_name[tok_type]
        if tok_type in ('NEWLINE', 'COMMENT', 'NL', 'INDENT', 'DEDENT'):
            continue
        elif tok_type == 'STRING':
            # cleandoc, not the raw string: a `""" Foo` opener would otherwise
            # indent the first line only, which rst reads as a block quote
            docstring = inspect.cleandoc(eval(tok_content))
            # If the docstring is formatted with several paragraphs, extract
            # the first one:
            paragraphs = '\n'.join(line.rstrip()
                                for line in docstring.split('\n')).split('\n\n')
            if len(paragraphs) > 0:
                first_par = paragraphs[0]
        break
    return docstring, first_par, erow+1+start_row


################################################################################
# class `ExampleLister`
################################################################################
class ExampleLister(object):
    """ Builds a rst-formatted list of examples from a list of files.
    """
    # Header template, for the example gallery.
    header_tpl = """
%(title)s
--------------------------------------------------

%(intro)s

.. toctree::
   :hidden:

%(toctree)s

.. A comment to split paragraphs

"""

    # Template used to create the example rst file
    example_rst_file_tpl = """

.. _example_%(short_file_name)s:

%(title)s example
--------------------------------------------

%(docstring)s

**Python source code:** :download:`%(short_file_name)s.py`

.. literalinclude:: %(short_file_name)s.py
    :lines: %(end_row)s-


    """

    # The title of the corresponding section in the example gallery.
    title = ''

    # The introductory text of the subsection
    intro =''

    def __init__(self, **kwargs):
        # Cheap unique hash for substitutions
        self._unique_hash = next(global_counter)
        for name, value in kwargs.items():
            setattr(self, name, value)


    def render_all(self, stream, file_list):
        """ Render the example list to the given
            stream (file-like object).
        """
        self._stream = stream
        files_details = self.render_header(file_list)
        for index, file_details in enumerate(files_details):
            filename, short_file_name, short_desc, title, docstring, \
                                                    end_row = file_details
            page = Path(self.out_dir, 'example_%s.rst' % short_file_name)
            with page.open('w') as stream:
                self.render_example_page(stream, index, file_details)
            self.gallery_entry(index, file_details)

        del self._stream


    def render_header(self, filenames):
        files_details = list()
        toctree = list()

        for filename in filenames:
            docstring, short_desc, end_row = extract_docstring(filename)
            short_file_name = os.path.basename(filename)[:-3]
            title = short_file_name.replace('_', ' ')
            title = title[0].upper() + title[1:]
            shutil.copy(filename,
                        os.path.join(self.out_dir, os.path.basename(filename)))


            toctree.append("""   example_%s.rst""" % short_file_name)
            files_details.append((filename, short_file_name, short_desc,
                                        title, docstring, end_row))

        toctree = '\n'.join(toctree)

        title = self.title
        intro = self.intro
        self._stream.write(self.header_tpl % locals())
        return files_details


    def render_example_page(self, stream, index, file_details):
        """ Render an individual example page.
        """
        filename, short_file_name, short_desc, title, docstring, end_row \
                                                                = file_details
        stream.write(self.example_rst_file_tpl % locals())


    def gallery_entry(self, index, file_details):
        """ Write the entry in the main example gallery file
            corresponding to the given file details.
        """
        filename, short_file_name, short_desc, title, docstring, \
                                end_row = file_details
        self._stream.write(
                "\n* :ref:`example_%(short_file_name)s`\n" % locals()
            )
        short_desc = short_desc.lstrip().rstrip()
        for line in short_desc.split('\n'):
            self._stream.write(4*" " + line.lstrip() + "\n")


################################################################################
# class `ImagesExampleLister`
################################################################################
class ImagesExampleLister(ExampleLister):
    """ ExampleLister that looks for thumbnails.
    """

    # Relative directory to images
    images_dir = 'mayavi/images/'

    def render_all(self, stream, file_list):
        self._stream = stream
        files_details = self.render_header(file_list)
        unique_hash = self._unique_hash
        for index, (filename, short_file_name, _, _, _, _) in \
                                        enumerate(files_details):
            image_file = os.path.join(self.images_dir,
                        'example_%(short_file_name)s.jpg' % locals())
            if os.path.exists(image_file):
                short_image_file = os.path.join(*(
                                    image_file.split(os.sep)[1:]))
                self._stream.write("""
.. |%(unique_hash)02i%(index)02i| image:: ../%(short_image_file)s
    :width: 150

            """ % locals())
            else:
                self._stream.write("""
.. |%(unique_hash)02i%(index)02i| raw:: html

    <br/>

            """ % locals())

        self._stream.write(2*('\n' + 7*"=" + " " + 45*"="))

        for index, file_details in enumerate(files_details):
            filename, short_file_name, short_desc, title, docstring, end_row = \
                                                                file_details
            page = Path(self.out_dir, 'example_%s.rst' % short_file_name)
            with page.open('w') as stream:
                self.render_example_page(stream, index, file_details)
            self.gallery_entry(index, file_details)

        self._stream.write("\n"+7*"=" + " " + 45*"=" + '\n')

        del self._stream


    def render_example_page(self, stream, index, file_details):
        """ Hijack this method to, optionally, render images.
        """
        # Jump one step up, and do not call ImagesExampleLister
        filename, short_file_name, short_desc, title, docstring, end_row = \
                                                                file_details
        image_file = os.path.join(self.images_dir,
                        'example_%(short_file_name)s.jpg' % locals())
        if os.path.exists(image_file):
            docstring += """

.. image:: ../%s
    :align: center

""" % os.path.join(*(image_file.split(os.sep)[1:]))

        file_details = \
            filename, short_file_name, short_desc, title, docstring, end_row

        stream.write(self.example_rst_file_tpl % locals())



    def gallery_entry(self, index, file_details):
        filename, short_file_name, short_desc, title, docstring, end_row = \
                                                                file_details
        # never break a word or a hyphenated one: a description mentioning
        # :ref:`data-structures-used-by-mayavi` would otherwise be split across
        # lines mid-label, leaving a reference that resolves to nothing
        short_desc = textwrap.wrap(short_desc, width=40,
                                   break_long_words=False,
                                   break_on_hyphens=False)
        unique_hash = self._unique_hash
        self._stream.write(
                ("\n|%(unique_hash)02i%(index)02i|" % locals()).ljust(9) +
                ":ref:`example_%(short_file_name)s`\n" % locals()
            )
        for line in short_desc:
            self._stream.write(9*" " + line.lstrip() + "\n")

################################################################################
# class `MlabExampleLister`
################################################################################
class RenderedExampleLister(ImagesExampleLister):
    """ ExampleLister that renders each example's figure rather than only
        looking for one already on disk.
    """

    render_images = False

    images_dir = 'mayavi/generated_images'

    def render_example_page(self, stream, index, file_details):
        """ Hijack this method to, optionally, render images.
        """
        filename, short_file_name, short_desc, title, docstring, end_row = \
                                                            file_details
        if self.render_images:
            image_file = os.path.join(self.images_dir, 'example_%s.jpg' \
                                    % short_file_name)
            capture_example(filename, short_file_name, image_file)
        ImagesExampleLister.render_example_page(self, stream,
                                                index, file_details)


################################################################################
# class `MlabExampleLister`
################################################################################
class MlabExampleLister(RenderedExampleLister):

    header_tpl = """
Mlab functions gallery
----------------------

These are the examples of the mlab plotting functions. They are
copied out here for convenience. Please refer to the corresponding
section of the user guide for more information (
:ref:`mlab_plotting_functions`).

.. currentmodule:: mayavi.mlab

+------------------+-------------------------+----------------------+
| :func:`plot3d`   | :func:`points3d`        | :func:`imshow`       |
|                  |                         |                      |
| |plot3d.jpg|     | |points3d.jpg|          | |imshow.jpg|         |
+------------------+-------------------------+----------------------+
| :func:`surf`     | :func:`contour_surf`    | :func:`mesh`         |
|                  |                         |                      |
| |surf.jpg|       | |contour_surf.jpg|      | |mesh.jpg|           |
+------------------+-------------------------+----------------------+
| :func:`barchart` | :func:`triangular_mesh` | :func:`contour3d`    |
|                  |                         |                      |
| |barchart.jpg|   | |triangular_mesh.jpg|   | |contour3d.jpg|      |
+------------------+-------------------------+----------------------+
| :func:`quiver3d` | :func:`flow`            | :func:`volume_slice` |
|                  |                         |                      |
| |quiver3d.jpg|   |  |flow.jpg|             | |volume_slice.jpg|   |
+------------------+-------------------------+----------------------+

.. |plot3d.jpg| image:: ../generated_images/mayavi_mlab_plot3d.jpg
     :width: 150

.. |points3d.jpg| image:: ../generated_images/mayavi_mlab_points3d.jpg
     :width: 150

.. |imshow.jpg| image:: ../generated_images/mayavi_mlab_imshow.jpg
     :width: 150

.. |contour_surf.jpg| image:: ../generated_images/mayavi_mlab_contour_surf.jpg
     :width: 150

.. |triangular_mesh.jpg| image:: ../generated_images/mayavi_mlab_triangular_mesh.jpg
     :width: 150

.. |surf.jpg| image:: ../generated_images/mayavi_mlab_surf.jpg
     :width: 150

.. |mesh.jpg| image:: ../generated_images/mayavi_mlab_mesh.jpg
     :width: 150

.. |barchart.jpg| image:: ../generated_images/mayavi_mlab_barchart.jpg
     :width: 150

.. |contour3d.jpg| image:: ../generated_images/mayavi_mlab_contour3d.jpg
     :width: 150

.. |quiver3d.jpg| image:: ../generated_images/mayavi_mlab_quiver3d.jpg
     :width: 150

.. |flow.jpg| image:: ../generated_images/mayavi_mlab_flow.jpg
     :width: 150

.. |volume_slice.jpg| image:: ../generated_images/mayavi_mlab_volume_slice.jpg
     :width: 150


Advanced mlab examples
-----------------------

.. toctree::
   :hidden:

%(toctree)s

"""

    example_rst_file_tpl = """

.. _example_%(short_file_name)s:

%(title)s example
--------------------------------------------------------------------

%(docstring)s

**Python source code:** :download:`%(short_file_name)s.py`

.. literalinclude:: %(short_file_name)s.py
    :lines: %(end_row)s-


    """




################################################################################
# Main entry point
def render_examples(render_images=False, out_dir='mayavi/auto'):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    example_gallery_file = StringIO()

    example_gallery_file.write("""

.. _example_gallery:

Example gallery
=================

""")

    ##########################################################################
    # Mlab examples
    example_files = [ filename
                    for filename in glob.glob(os.path.join(EXAMPLE_DIR,
                        'mlab', '*.py'))
                    if is_mlab_example(filename)]
    # Sort by file length (gives a measure of the complexity of the
    # example)
    example_files.sort(key=lambda name: (len(Path(name).read_text().splitlines()), name))

    mlab_example_lister = MlabExampleLister(render_images=render_images,
                                        out_dir=out_dir,
                                        images_dir='mayavi/generated_images')
    if render_images:
        pass
        # XXX: Add logics to deal with rerendering examples cleverly

    mlab_example_lister.render_all(example_gallery_file, example_files)

    ##########################################################################
    # Interactive application examples
    example_files = [ filename
                    for filename in glob.glob(os.path.join(EXAMPLE_DIR,
                        'interactive', '*.py'))]
    # Sort by file length (gives a measure of the complexity of the
    # example)
    example_files.sort(key=lambda name: (len(Path(name).read_text().splitlines()), name))
    example_lister = RenderedExampleLister(
            render_images=render_images,
            images_dir='mayavi/generated_images',
            title="Interactive examples",
            out_dir=out_dir,
            intro="""

Examples showing how to use the interactive features of Mayavi, either
via the mayavi2 application, or via specially-crafted dialogs and
applications.
    """)
    example_lister.render_all(example_gallery_file, example_files)

    ##########################################################################
    # Advanced visualization examples
    example_files = [ filename
                    for filename in glob.glob(os.path.join(EXAMPLE_DIR,
                        'advanced_visualization', '*.py'))]
    # Sort by file length (gives a measure of the complexity of the
    # example)
    example_files.sort(key=lambda name: (len(Path(name).read_text().splitlines()), name))
    example_lister = RenderedExampleLister(
            render_images=render_images,
            images_dir='mayavi/generated_images',
            title="Advanced visualization examples",
            out_dir=out_dir,
            intro="""
Data visualization using the core Mayavi API, object-oriented, and with
more fine control than mlab.

    """)
    example_lister.render_all(example_gallery_file, example_files)

    ##########################################################################
    # Data interaction examples
    example_files = [ filename
                    for filename in glob.glob(os.path.join(EXAMPLE_DIR,
                        'data_interaction', '*.py'))]
    # Sort by file length (gives a measure of the complexity of the
    # example)
    example_files.sort(key=lambda name: (len(Path(name).read_text().splitlines()), name))
    example_lister = RenderedExampleLister(
            render_images=render_images,
            images_dir='mayavi/generated_images',
            title="Data interaction examples",
            out_dir=out_dir,
            intro="""
Examples showing how you can query and interact with the data.

    """)
    example_lister.render_all(example_gallery_file, example_files)

    ##########################################################################
    # The remaining files
    example_files = [ filename
                    for filename in glob.glob(os.path.join(EXAMPLE_DIR,
                        '*.py'))]
    # Sort by file length (gives a measure of the complexity of the
    # example)
    example_files.sort(key=lambda name: (len(Path(name).read_text().splitlines()), name))
    example_lister = ExampleLister(title="Misc examples",
                                   out_dir=out_dir)
    example_lister.render_all(example_gallery_file, example_files)


    Path(out_dir, 'examples.rst').write_text(example_gallery_file.getvalue())


if __name__ == '__main__':
    render_examples()



import unittest
from unittest.mock import patch

from mayavi.action import help as action_help


class TestBrowserOpen(unittest.TestCase):
    def test_opens_url(self):
        url = '/tmp/Mayavi Docs; unexpected-command'
        with patch.object(action_help.webbrowser, 'open') as open_mock:
            action_help.browser_open(url)

        open_mock.assert_called_once_with(url, autoraise=1)


if __name__ == '__main__':
    unittest.main()



""" Tests for the garbage collection of TVTKScene objects.

"""
# Authors: Deepak Surti, Ioannis Tziakos
# Copyright (c) Enthought, Inc.
# License: BSD Style.

import sys
import unittest
import weakref
import gc

from traits.etsconfig.api import ETSConfig
from tvtk.pyface.tvtk_scene import TVTKScene
from tvtk.tests.common import restore_gc_state
from tvtk.vtk_module import vtkRenderWindow


class TestTVTKScene(unittest.TestCase):

    @unittest.skipIf(
        sys.platform.startswith('win') or ETSConfig.toolkit == 'null',
        'CI with windows fails due to lack of OpenGL, or toolkit is null.'
    )
    def test_offscreen_render_window_left_to_vtk(self):
        # An offscreen scene must use whatever concrete render window VTK's
        # own factory picks, and no other.  We used to instantiate
        # EGLRenderWindow whenever the class existed, which since the wheels
        # started shipping EGL support means always -- and segfaulted on every
        # machine with no EGL driver behind it.  See
        # https://github.com/enthought/mayavi/issues/1332
        scene = TVTKScene(off_screen_rendering=True)
        try:
            renwin = scene.render_window
            self.assertEqual(
                renwin.class_name, vtkRenderWindow().GetClassName()
            )
            self.assertTrue(renwin.off_screen_rendering)
        finally:
            scene.close()

    @unittest.skipIf(
        sys.platform.startswith('win') or ETSConfig.toolkit == 'null',
        'CI with windows fails due to lack of OpenGL, or toolkit is null.'
    )
    def test_tvtk_scene_garbage_collected(self):

        # given
        scene_collected = []
        scene_weakref = None

        def scene_collected_callback(weakref):
            scene_collected.append(True)

        def do():
            scene = TVTKScene()
            reference = weakref.ref(scene, scene_collected_callback)
            scene.close()
            return reference

        # when
        with restore_gc_state():
            gc.disable()
            scene_weakref = do()

        # The TVTK Scene should have been collected.
        self.assertTrue(scene_collected[0])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Setup script for TVTK, numpy.distutils based.
#
#

import os
import sys

from setuptools import Distribution


def can_compile_extensions():
    try:
        import numpy  # noqa
    except Exception:
        return False  # our extension needs numpy/arrayobject.h
    sargs = {'script_name': None, 'script_args': ["--build-ext"]}
    d = Distribution(sargs)
    cfg = d.get_command_obj('config')
    cfg.dump_source = 0
    cfg.noisy = 0
    cfg.finalize_options()
    build_ext = d.get_command_obj('build_ext')
    build_ext.finalize_options()
    try:
        result = cfg.try_compile(
            'int main(void) {return 0;}',
            headers=['Python.h'],
            include_dirs=build_ext.include_dirs,
            lang='c'
        )
    except Exception as exc:
        print(f"Compilation failed, assuming no C compiler: {exc}")
        return False
    else:
        return result


def gen_tvtk_classes_zip():
    MY_DIR = os.path.dirname(__file__)
    sys.path.append(MY_DIR)
    from tvtk.code_gen import TVTKGenerator
    target = os.path.join(MY_DIR, 'tvtk_classes.zip')
    output_dir = os.path.dirname(target)
    try:
        os.mkdir(output_dir)
    except:
        pass
    print('-'*70)
    if os.path.exists(target):
        print('Deleting possibly old TVTK classes')
        os.unlink(target)
    print("Building TVTK classes...", end=' ')
    sys.stdout.flush()
    cwd = os.getcwd()
    os.chdir(output_dir)
    gen = TVTKGenerator('')
    gen.generate_code()
    gen.build_zip(True)
    os.chdir(cwd)
    print("Done.")
    print('-'*70)
    sys.path.remove(MY_DIR)


def vtk_version_changed(zipfile):
    """Checks the ZIP file's VTK build version versus the current
    installed version of VTK and returns `True` if the versions are
    different.

    """
    result = True
    if os.path.exists(zipfile):
        import vtk
        vtk_version = vtk.vtkVersion().GetVTKVersion()[:3]
        sys.path.append(zipfile)
        try:
            from tvtk_classes.vtk_version import vtk_build_version
        except ImportError:
            result = True
        else:
            if vtk_version != vtk_build_version:
                result = True
            else:
                result = False
        sys.path.pop()

    return result



"""
Retrieve radar data from the NASA and plot a view of the Grand Canyon
landscape.

We cannot display the whole data, as it would be too big. To display
more, see the canyon decimation example.

This example is interesting as it shows how numpy can be used to load
and crop data completly foreign to Mayavi.
"""
# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Retrieve the grand Canyon topological data ##################################
import os
# Original file:
#'ftp://e0srp01u.ecs.nasa.gov/srtm/version2/SRTM1/Region_04/N36W113.hgt.zip'
if not os.path.exists('N36W113.hgt.zip'):
    # Download the data
    try:
        from urllib import urlopen
    except ImportError:
        from urllib.request import urlopen
    print('Downloading data, please wait (10M)')
    opener = urlopen(
    'https://s3.amazonaws.com/storage.enthought.com/www/sample_data/N36W113.hgt.zip'
        )
    open('N36W113.hgt.zip', 'wb').write(opener.read())

# Load the data (signed 2 byte integers, big endian) ##########################
import zipfile
import numpy as np

data = np.frombuffer(zipfile.ZipFile('N36W113.hgt.zip').read('N36W113.hgt'),
                    '>i2')
data = data.reshape((3601, 3601))
data = data.astype(np.float32)

# Plot an interesting section #################################################
from mayavi import mlab
data = data[:1000, 900:1900]
# Convert missing values into something more sensible.
data[data == -32768] = data[data > 0].min()

mlab.figure(size=(400, 320), bgcolor=(0.16, 0.28, 0.46))
mlab.surf(data, colormap='gist_earth', warp_scale=0.2,
            vmin=1200, vmax=1610)
# The data takes a lot of memory, and the surf command has created a
# copy. We free the inital memory.
del data

# A view of the canyon
mlab.view(-5.9, 83, 570, [5.3, 20, 238])
mlab.show()



"""
A custom dialog to adjust the parameters of a GeometryFilter to crop
data points.

This example shows how to use a GeometryFilter to crop data points, but
also how to build a custom dialog to easily set interactively parameters
of a filter, or any other Mayavi object.

The GeometryFilter crops all data within a bounding box specified by
'extents'. In this example, we want to be able to tweak these extents
interactively. For this, we build a Traits object that has 'x_min',
'x_max', 'y_min', ... attributes. Traits enables us to represent this
object as a dialog box. We use a callback called when these attributes
are modified to propagate them to the filter. For more information
on creating GUIs with Traits:

    https://docs.enthought.com/traitsui

    https://docs.enthought.com/traits

    https://support.enthought.com/hc/en-us/articles/204469620-Introductory-materials-for-Traits-and-Traits-UI

"""
# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.


################################################################################
# Create a set of points, with given density
import numpy as np
x, y, z, s = np.random.random((4, 500))

################################################################################
# A dialog to edit a range interactively and propagate it to the filter
from traits.api import HasTraits, Range, Float, Instance, \
    on_trait_change
from traitsui.api import View

class ExtentDialog(HasTraits):
    """ A dialog to graphical adjust the extents of a filter.
    """

    # Data extents
    data_x_min = Float
    data_x_max = Float
    data_y_min = Float
    data_y_max = Float
    data_z_min = Float
    data_z_max = Float

    x_min = Range('data_x_min', 'data_x_max', 'data_x_min')
    x_max = Range('data_x_min', 'data_x_max', 'data_x_max')
    y_min = Range('data_y_min', 'data_y_max', 'data_y_min')
    y_max = Range('data_y_min', 'data_y_max', 'data_y_max')
    z_min = Range('data_z_min', 'data_z_max', 'data_z_min')
    z_max = Range('data_z_min', 'data_z_max', 'data_z_max')

    filter = Instance(HasTraits, allow_none=False)

    @on_trait_change('x_min,x_max,y_min,y_max,z_min,z_max')
    def update_extent(self):
        if (self.filter is not None
                    and self.x_min < self.x_max
                    and self.y_min < self.y_max
                    and self.z_min < self.z_max
                            ):
            self.filter.extent = (self.x_min, self.x_max,
                                  self.y_min, self.y_max,
                                  self.z_min, self.z_max)

    view = View('x_min', 'x_max', 'y_min', 'y_max', 'z_min', 'z_max',
                title='Edit extent', resizable=True)


################################################################################
# Now build the visualization using mlab
from mayavi import mlab
fig = mlab.figure(1, bgcolor=(1, 1, 1))

# Create unconnected points
pts = mlab.pipeline.scalar_scatter(x, y, z, s)
mlab.outline(pts)

# Use a geometry_filter to filter with a bounding box
geometry_filter = mlab.pipeline.user_defined(pts,
                                   filter='GeometryFilter')
geometry_filter.filter.extent_clipping = True
# Connect our dialog to the filter
extent_dialog = ExtentDialog(
            data_x_min=0, data_x_max=1,
            data_y_min=0, data_y_max=1,
            data_z_min=0, data_z_max=1,
            filter=geometry_filter.filter)
# We need to use 'edit_traits' and not 'configure_traits()' as we do
# not want to start the GUI event loop (the call to mlab.show())
# at the end of the script will do it.
extent_dialog.edit_traits()

# The geometry_filter leaves hanging points, we need to add a
# CleanPolyData filter to get rid of these.
clip = mlab.pipeline.user_defined(geometry_filter,
                                    filter='CleanPolyData')

# Finally, visualize the remaining points with spheres using a glyph
# module
spheres = mlab.pipeline.glyph(clip, scale_factor=.1)

mlab.show()



#!/usr/bin/env python

""" This is a simple test envisage3 application to demonstrate how one
can use TVTK's scene and browser plugins to create a simple application.
"""

# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import os.path
import logging

# Enthought library imports.
from apptools.logger.api import LogFileHandler, FORMATTER
from traits.etsconfig.api import ETSConfig
from traits.api import List
from envisage.api import Plugin
from envisage.ui.workbench.api import WorkbenchApplication
from envisage.core_plugin import CorePlugin
from envisage.ui.workbench.workbench_plugin import WorkbenchPlugin
from envisage.plugins.python_shell.python_shell_plugin import PythonShellPlugin
from tvtk.plugins.scene.scene_plugin import ScenePlugin
from tvtk.plugins.scene.ui.scene_ui_plugin import SceneUIPlugin
from tvtk.plugins.browser.browser_plugin import BrowserPlugin
from pyface.workbench.api import Perspective, PerspectiveItem


logger = logging.getLogger()

###############################################################################
# `TestPerspective` class.
###############################################################################
class TestPerspective(Perspective):
    """ An default perspective for the app. """

    # The perspective's name.
    name = 'Test'

    # Should this perspective be enabled or not?
    enabled = True

    # Should the editor area be shown in this perspective?
    show_editor_area = True

    # View IDs.
    BROWSER_VIEW = 'tvtk.plugins.browser.browser_view.BrowserView'
    SHELL_VIEW = 'envisage.plugins.python_shell.view.python_shell_view.PythonShellView'

    # The contents of the perspective.
    contents = [
        PerspectiveItem(id=BROWSER_VIEW, position='left'),
        PerspectiveItem(id=SHELL_VIEW, position='bottom')
    ]

###############################################################################
# `MyPlugin` class.
###############################################################################
class MyPlugin(Plugin):

    # Extension points we contribute to.
    PERSPECTIVES = 'envisage.ui.workbench.perspectives'

    # The plugin's unique identifier.
    id = 'tvtk_example_plugin'

    # The plugin's name (suitable for displaying to the user).
    name = 'TVTK example plugin'

    # Perspectives.
    perspectives = List(contributes_to=PERSPECTIVES)

    def _perspectives_default(self):
        """ Trait initializer. """
        return [TestPerspective]


###############################################################################
# Useful functions.
###############################################################################
def setup_logger(logger, fname, stream=True, mode=logging.ERROR):
    """Setup a log file and the logger in `ETSConfig.application_home`.

    Parameters:
    -----------

    fname -- file name the logger should use.

    stream -- Add a stream handler.

    mode -- the logging mode.

    """
    path = os.path.join(ETSConfig.application_home, fname)
    handler = LogFileHandler(path)
    logger.addHandler(handler)
    if stream:
        s = logging.StreamHandler()
        s.setFormatter(FORMATTER)
        s.setLevel(mode)
        logger.addHandler(s)
    logger.info("*"*80)
    logger.info("logfile is: '%s'", os.path.abspath(path))
    logger.info("*"*80)


def bind_object(value, app):
    """Binds the scene manager service to the Python shell."""
    if not value:
        # value is False when the GUI is stopped.
        return
    id = 'envisage.plugins.python_shell.view.python_shell_view.PythonShellView'
    py = app.workbench.active_window.get_view_by_id(id)

    id = 'tvtk.plugins.scene.i_scene_manager.ISceneManager'
    sm = app.workbench.active_window.get_service(id)
    if py is not None:
        py.bind('scene_manager', sm)


def main():
    """The main application is created and launched here."""
    # Setup the logger.

    plugins = [CorePlugin(),
               WorkbenchPlugin(),
               MyPlugin(),
               ScenePlugin(),
               SceneUIPlugin(),
               BrowserPlugin(),
               PythonShellPlugin(),
               ]
    # Create an Envisage application.
    id = 'tvtk.examples.plugins.test'
    application = WorkbenchApplication(id=id,
                                       plugins = plugins
                                       )
    # This needs to be done here since the ETSConfig.application_home is
    # not set correctly up to this point.
    setup_logger(logger, 'test.log', mode=logging.DEBUG)

    application.gui.on_trait_change(lambda value: bind_object(value, application),
                                    'started')
    # Start the application.
    application.run()


if __name__ == '__main__':
    main()




"""
An example which shows how to plot a surface from data acquired
irregularly.

Data giving the variation of a parameter 'z' as a function of two others
('x' and 'y') is often plotted as a `carpet plot`, using a surface to
visualize the underlying function. when the data has been acquired on a
regular grid for parameters 'x' and 'y', it can simply be view with the
mlab.surf function. However, when there are some missing points, or the
data has been acquired at random, the surf function cannot be used.

The difficulty stems from the fact that points positioned in 3D do
not define a surface if no connectivity information is given. With the
surf function, this information is implicit from the shape of the input
arrays.

In this example, randomly-positioned points in the (x, y) plane are
embedded in a surface in the z axis. We first visualize the points using
mlab.points3d. We then use the delaunay2d filter to extract the mesh by
nearest-neighboor matching, and visualize it using the surface module.
"""
# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

import numpy as np

# Create data with x and y random in the [-2, 2] segment, and z a
# Gaussian function of x and y.
np.random.seed(12345)
x = 4 * (np.random.random(500) - 0.5)
y = 4 * (np.random.random(500) - 0.5)


def f(x, y):
    return np.exp(-(x ** 2 + y ** 2))

z = f(x, y)

from mayavi import mlab
mlab.figure(1, fgcolor=(0, 0, 0), bgcolor=(1, 1, 1))

# Visualize the points
pts = mlab.points3d(x, y, z, z, scale_mode='none', scale_factor=0.2)

# Create and visualize the mesh
mesh = mlab.pipeline.delaunay2d(pts)
surf = mlab.pipeline.surface(mesh)

mlab.view(47, 57, 8.2, (0.1, 0.15, 0.14))
mlab.show()



#!/usr/bin/env mayavi2
"""This script demonstrates how one can script Mayavi's core API to display
streamlines and an iso surface.
"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Standard library imports
from os.path import join, abspath, dirname

# Enthought library imports
from mayavi.scripts import mayavi2
from mayavi.sources.vtk_xml_file_reader import VTKXMLFileReader
from mayavi.sources.vrml_importer import VRMLImporter
from mayavi.modules.outline import Outline
from mayavi.modules.streamline import Streamline
from mayavi.modules.iso_surface import IsoSurface

def setup_data(fname):
    """Given a VTK XML file name `fname`, this creates a mayavi2
    reader for it and adds it to the pipeline.  It returns the reader
    created.
    """
    r = VTKXMLFileReader()
    r.initialize(fname)
    mayavi.add_source(r)
    return r

def show_vrml(fname):
    """Given a VRML file name it imports it into the scene.
    """
    r = VRMLImporter()
    r.initialize(fname)
    mayavi.add_source(r)
    return r

def streamline():
    """Sets up the mayavi pipeline for the visualization.
    """
    # Create an outline for the data.
    o = Outline()
    mayavi.add_module(o)

    s = Streamline(streamline_type='tube')
    mayavi.add_module(s)
    s.stream_tracer.integration_direction = 'both'
    s.seed.widget.center = 3.5, 0.625, 1.25
    s.module_manager.scalar_lut_manager.show_scalar_bar = True

    i = IsoSurface()
    mayavi.add_module(i)
    i.contour.contours[0] = 550
    i.actor.property.opacity = 0.5

@mayavi2.standalone
def main():
    mayavi.new_scene()

    data_dir = mayavi2.get_data_dir(dirname(abspath(__file__)))
    vrml_fname = join(data_dir, 'room_vis.wrl')
    r = show_vrml(vrml_fname)

    fname = join(data_dir, 'fire_ug.vtu')
    r = setup_data(fname)
    streamline()


if __name__ == '__main__':
    main()



"""Keep Qt off the wayland platform, where VTK render windows cannot embed.

VTK has no Wayland render window: on Unix it draws into a native X11 window
looked up by the id Qt reports.  Qt's wayland backend fabricates ``winId()``
values with no X11 window behind them, so VTK's first X request on one fails
with BadWindow and Xlib aborts the process (see tvtk/WORKAROUNDS.md and
enthought/mayavi#1396).  The fix is to run Qt on XWayland instead, which is
what ``QT_QPA_PLATFORM=xcb`` asks for.

``steer_qt_to_x11`` applies that while the choice is still open -- Qt reads
``QT_QPA_PLATFORM`` when the QApplication is constructed -- so it has to run
before pyface's toolkit initialization, which creates the QApplication as an
import side effect.  ``mayavi/__init__.py`` and ``tvtk/pyface/__init__.py``
call it for that reason.  For the cases it cannot win (the user imported
``traitsui``/``pyface`` before anything mayavi, or pinned
``QT_QPA_PLATFORM=wayland`` themselves), ``embedding_error`` gives the Qt
scene widget a clear error to raise in place of the Xlib abort.

This module must stay importable without Qt, VTK, or a built tvtk: it runs
at ``import mayavi``.
"""
# Author: Enthought, Inc.
# License: BSD Style.

import os


def x11_platform_override(environ):
    """The Qt platform to request so VTK render windows can embed, or None.

    Keyed off the environment rather than ``sys.platform``, so that every OS
    that can run a Wayland session (Linux, the BSDs) is covered without
    naming them: ``WAYLAND_DISPLAY`` is what marks one.

    Pure logic split out for testing; see the module docstring for why.
    """
    if environ.get('ETS_TOOLKIT', 'qt4') not in ('qt4', 'qt', ''):
        return None  # wx or null toolkit: leave Qt alone
    if 'QT_QPA_PLATFORM' in environ:  # an explicit choice wins
        return None
    if 'WAYLAND_DISPLAY' not in environ:  # not a Wayland session
        return None
    if 'DISPLAY' not in environ:  # no X server to steer onto
        return None
    return 'xcb'


def steer_qt_to_x11():
    """Set QT_QPA_PLATFORM=xcb where a wayland default would break VTK."""
    override = x11_platform_override(os.environ)
    if override is not None:
        os.environ['QT_QPA_PLATFORM'] = override


def embedding_error(platform_name):
    """Why a VTK render window cannot embed in a Qt window here, or None.

    ``platform_name`` is ``QGuiApplication.platformName()`` of the running
    application -- the OS is left out of it for the reason
    ``x11_platform_override`` leaves it out -- so this can only be checked,
    and the clear error raised in place of the Xlib abort, once steering is
    already too late.
    """
    if platform_name.startswith('wayland'):
        return (
            "Mayavi cannot embed a VTK render window in Qt on Wayland (Qt is "
            "running on the '%s' platform).  Set the environment variable "
            "QT_QPA_PLATFORM=xcb and try again." % platform_name)
    return None



"""Test for the hide-show feature.
"""
# Author: Prabhu Ramachandran <prabhu [at] aero . iitb . ac . in>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Standard library imports.
from os.path import abspath
from io import BytesIO
import copy

# Local imports.
from common import TestCase, get_example_data


class TestHideShow(TestCase):

    def check(self, saved=False):
        """Does the checking, if saved is True it does not change the
        properties at first to see how those behave and only tests the
        final unpickled state."""
        script = self.script
        e = script.engine
        scene = e.current_scene
        wrl = scene.children[0]
        src = scene.children[1]
        mm = src.children[0]
        scp = mm.children[0]
        iso = mm.children[1]
        # Every check below this that involves the VRML source is a loop over
        # its actors, so an import that quietly brought nothing in would leave
        # them all vacuously true -- which is exactly how the source calling a
        # method VTK 9.6 had removed went unnoticed.
        assert len(wrl.actors) > 0
        if not saved:
            assert scp.actor.actor.visibility == True
            assert scp.implicit_plane.widget.enabled == True
            for a in wrl.actors:
                assert a.visibility == True

            assert iso.actor.actor.visibility == True

            # Check if widget state is remembered.
            scp.implicit_plane.widget.enabled = False
            scp.visible = False
            assert scp.actor.actor.visibility == False
            assert scp.implicit_plane.widget.enabled == False
            assert scp.name == 'ScalarCutPlane [Hidden]'
            # Reenable it and check widget state.
            scp.visible = True
            assert scp.actor.actor.visibility == True
            assert scp.implicit_plane.widget.enabled == False

            # Reset the visible state.
            wrl.visible = False
            scp.visible = False
            iso.visible = False

        # Check final state.
        for a in wrl.actors:
            assert a.visibility == False
        assert wrl.name.find('[Hidden]') > -1

        assert scp.actor.actor.visibility == False
        assert scp.implicit_plane.widget.enabled == False
        assert scp.name == 'ScalarCutPlane [Hidden]'
        assert iso.name == 'IsoSurface [Hidden]'
        assert iso.actor.actor.visibility == False


    def test(self):
        self.main()

    def do(self):
        ############################################################
        # Imports.
        from mayavi.sources.api import VTKXMLFileReader,\
                VRMLImporter
        from mayavi.modules.api import ScalarCutPlane,\
                IsoSurface

        ############################################################
        # Create a new scene and set up the visualization.
        s = self.new_scene()
        script = mayavi = self.script

        # Read a VRML file.
        w = VRMLImporter()
        w.initialize(get_example_data('room_vis.wrl'))
        script.add_source(w)

        # Read a VTK data file.
        r = VTKXMLFileReader()
        r.initialize(get_example_data('fire_ug.vtu'))
        script.add_source(r)

        # Create the modules.
        scp = ScalarCutPlane()
        script.add_module(scp)

        iso = IsoSurface()
        script.add_module(iso)

        # Check.
        self.check(saved=False)

        ############################################################
        # Test if saving a visualization and restoring it works.

        # Save visualization.
        f = BytesIO()
        f.name = abspath('test.mv2') # We simulate a file.
        script.save_visualization(f)
        f.seek(0) # So we can read this saved data.

        # Remove existing scene.
        engine = script.engine
        engine.close_scene(s)

        # Load visualization
        script.load_visualization(f)
        s = engine.current_scene
        s.scene.isometric_view()

        # Now do the check.
        self.check(saved=True)

        ############################################################
        # Test if the Mayavi2 visualization can be deep-copied.

        # Pop the source object.
        sources = s.children
        s.children = []
        # Add it back to see if that works without error.
        s.children.extend(sources)
        # Now do the check.
        s.scene.isometric_view()
        self.check(saved=True)

        # Now deepcopy the source and replace the existing one with
        # the copy.  This basically simulates cutting/copying the
        # object from the UI via the right-click menu on the tree
        # view, and pasting the copy back.
        sources1 = copy.deepcopy(sources)
        s.children[:] = sources1
        # Now do the check.
        s.scene.isometric_view()
        self.check(saved=True)

        # If we have come this far, we are golden!

if __name__ == "__main__":
    t = TestHideShow()
    t.test()




"""Regression test for mlab.savefig(size=...) on an onscreen figure (gh-1288).

The first savefig used to capture at the window's size instead of the
requested one: set_size resized the embedded scene widget, which the
enclosing layout immediately reasserted, and the pre-layout widget size
inflated the auto magnification.
"""
# Copyright (c) Enthought, Inc.
# License: BSD Style.

import os
import shutil
import subprocess
import sys

import pytest

_SCRIPT = """
import os
import tempfile

from mayavi import mlab
from tvtk.api import tvtk


def png_size(fname):
    r = tvtk.PNGReader(file_name=fname)
    r.update()
    ext = r.output.extent
    return (ext[1] + 1, ext[3] + 1)


d = tempfile.mkdtemp()
# explicit, since both cases below are relative to it: (100, 80) is smaller
# than the window and (1200, 900) larger, exercising the magnification path
mlab.figure(size=(400, 350))
sizes = []
for n in range(3):
    mlab.test_contour3d()
    mlab.draw()
    fname = os.path.join(d, 'i%d.png' % n)
    mlab.savefig(fname, size=(100, 80))
    mlab.clf()
    sizes.append(png_size(fname))
# the first save is the regression: it saw the unsettled window size
assert sizes == [(100, 80)] * 3, sizes

# larger than the window exercises the magnification path
big = os.path.join(d, 'big.png')
mlab.savefig(big, size=(1200, 900))
assert png_size(big) == (1200, 900), png_size(big)
print('OK')
"""


@pytest.mark.timeout(180)
def test_savefig_size_onscreen():
    if sys.platform != 'linux':
        pytest.skip('needs xvfb-run or an X display; exercised on Linux')
    from traits.etsconfig.api import ETSConfig
    if ETSConfig.toolkit not in ('', 'qt', 'qt4'):
        pytest.skip(f'Qt-only test, got toolkit={ETSConfig.toolkit!r}')
    cmd = [sys.executable, '-c', _SCRIPT]
    xvfb = shutil.which('xvfb-run')
    if xvfb is not None:
        # a fresh server keeps windows off the developer's desktop, and
        # nests fine inside CI's own Xvfb
        cmd = [xvfb, '-a'] + cmd
    elif not os.environ.get('DISPLAY'):
        pytest.skip('no display and no xvfb-run')
    res = subprocess.run(cmd, capture_output=True, text=True, timeout=150)
    assert res.returncode == 0, res.stdout + res.stderr
    assert 'OK' in res.stdout, res.stdout



from mayavi import mlab
from pyface.api import GUI

def close():
    """Close the scene."""
    f = mlab.gcf()
    e = mlab.get_engine()
    e.window.workbench.prompt_on_exit = False
    e.window.close()
    mlab.options.backend = 'auto'
    # Hack: on Linux the splash screen does not go away so we force it.
    GUI.invoke_after(500, e.window.workbench.application.gui.stop_event_loop)

def test_mlab_envisage():
    "Test if mlab runs correctly when the backend is set to 'envisage'."
    @mlab.show
    def f():
        from mayavi.preferences.api import preference_manager
        preference_manager.root.show_splash_screen = False
        mlab.options.backend = 'envisage'
        mlab.test_contour3d()
        GUI.invoke_after(3000, close)

    f()

if __name__ == '__main__':
    test_mlab_envisage()



#!/usr/bin/env python
"""
A very simple example to show how you can use TVTK's visual module
with mlab and create simple animations.

In the example, the grey box bounces back and forth between the two red
ones.

The `tvtk.tools.visual` module exposes an API similar to
`VPython <http://vpython.org/>`_ and is useful to create animation based
on rigid object movement.

The `@animate` decorator ( :func:`mayavi.mlab.animate` ) is
detailed on section :ref:`animating_a_visualization`.

If you want to modify the data plotted by the mlab (as in the
`mlab.test_plot3d()` call) to create an animation, please see section
:ref:`mlab-animating-data`.

"""
# Author: Prabhu Ramachandran <prabhu [at] aero.iitb.ac.in>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

from mayavi import mlab
from tvtk.tools import visual
# Create a figure
f = mlab.figure(size=(500,500))
# Tell visual to use this as the viewer.
visual.set_viewer(f)

# A silly visualization.
mlab.test_plot3d()

# Even sillier animation.
b1 = visual.box()
b2 = visual.box(x=4., color=visual.color.red)
b3 = visual.box(x=-4, color=visual.color.red)
b1.v = 5.0

@mlab.show
@mlab.animate(delay=250)
def anim():
    """Animate the b1 box."""
    while 1:
        b1.x = b1.x + b1.v*0.1
        if b1.x > 2.5 or b1.x < -2.5:
            b1.v = -b1.v
        yield

# Run the animation.
anim()



"""Pytest configuration for the tvtk test suite."""
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# One filter per line, "#" comments allowed.
WARNING_LINES = r"""
error::
# these tests instantiate every VTK class and read every getter, deprecated
# ones included; the parenthetical wording varies far too much to match on
ignore:Call to deprecated .*Deprecated since version.*:DeprecationWarning
# should be fixed in traits
ignore: module 'sre_.+' is deprecated:DeprecationWarning
"""


def pytest_configure(config):
    """Add the warning filters this suite needs."""
    # Later entries win, so skip any the other suite's conftest already added:
    # a second "error::" would outrank the ignores it put there (`pytest mayavi
    # tvtk` runs both, and pluggy calls the two hooks in reverse load order).
    existing = config.getini('filterwarnings')
    for line in WARNING_LINES.split('\n'):
        line = line.strip()
        if line and not line.startswith('#') and line not in existing:
            config.addinivalue_line('filterwarnings', line)



#!/usr/bin/env python
"""
This script demonstrates how one can script the Mayavi application by
subclassing the application, create a new scene and create a few
simple modules.

This should be run as::

   $ python test.py

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Standard library imports
from os.path import join, abspath, dirname

# Enthought library imports
from mayavi.plugins.app import Mayavi
from mayavi.scripts.util import get_data_dir

class MyApp(Mayavi):
    def run(self):
        """This is executed once the application GUI has started.
        *Make sure all other MayaVi specific imports are made here!*
        """
        # Various imports to do different things.
        from mayavi.sources.vtk_file_reader import VTKFileReader
        from mayavi.modules.outline import Outline
        from mayavi.modules.axes import Axes
        from mayavi.modules.grid_plane import GridPlane
        from mayavi.modules.image_plane_widget import ImagePlaneWidget
        from mayavi.modules.text import Text

        script = self.script
        # Create a new scene.
        script.new_scene()

        # Read a VTK (old style) data file.
        r = VTKFileReader()
        r.initialize(join(get_data_dir(dirname(abspath(__file__))), 'heart.vtk'))
        script.add_source(r)

        # Put up some text.
        t = Text(text='MayaVi rules!', x_position=0.2,
                 y_position=0.9, width=0.8)
        t.property.color = 1, 1, 0  # Bright yellow, yeah!
        script.add_module(t)

        # Create an outline for the data.
        o = Outline()
        script.add_module(o)

        # Create an axes for the data.
        a = Axes()
        script.add_module(a)

        # Create an orientation axes for the scene.  This only works with
        # VTK-4.5 and above which is why we have the try block.
        try:
            from mayavi.modules.orientation_axes import OrientationAxes
        except ImportError:
            pass
        else:
            a = OrientationAxes()
            a.marker.set_viewport(0.0, 0.8, 0.2, 1.0)
            script.add_module(a)

        # Create three simple grid plane modules.
        # First normal to 'x' axis.
        gp = GridPlane()
        script.add_module(gp)
        # Second normal to 'y' axis.
        gp = GridPlane()
        gp.grid_plane.axis = 'y'
        script.add_module(gp)
        # Third normal to 'z' axis.
        gp = GridPlane()
        script.add_module(gp)
        gp.grid_plane.axis = 'z'

        # Create one ImagePlaneWidget.
        ipw = ImagePlaneWidget()
        script.add_module(ipw)
        # Set the position to the middle of the data.
        ipw.ipw.slice_position = 16



if __name__ == '__main__':
    a = MyApp()
    a.main()



"""
Plot spherical harmonics on the surface of the sphere, as well as a 3D
polar plot.

This example requires scipy.

In this example we use the mlab's mesh function:
:func:`mayavi.mlab.mesh`.
For plotting surfaces this is a very versatile function. The surfaces can
be defined as functions of a 2D grid.

For each spherical harmonic, we plot its value on the surface of a
sphere, and then in polar. The polar plot is simply obtained by varying
the radius of the previous sphere.
"""

# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

from mayavi import mlab
import numpy as np
from scipy.special import sph_harm_y

# Create a sphere
r = 0.3
pi = np.pi
cos = np.cos
sin = np.sin
phi, theta = np.mgrid[0:pi:101j, 0:2 * pi:101j]

x = r * sin(phi) * cos(theta)
y = r * sin(phi) * sin(theta)
z = r * cos(phi)

mlab.figure(1, bgcolor=(1, 1, 1), fgcolor=(0, 0, 0), size=(400, 300))
mlab.clf()
# Represent spherical harmonics on the surface of the sphere
for n in range(1, 6):
    for m in range(n):
        # sph_harm_y takes the polar angle first; the removed sph_harm(m, n,
        # theta, phi) took the azimuthal one
        s = sph_harm_y(n, m, phi, theta).real

        mlab.mesh(x - m, y - n, z, scalars=s, colormap='jet')

        s[s < 0] *= 0.97

        s /= s.max()
        mlab.mesh(s * x - m, s * y - n, s * z + 1.3,
                  scalars=s, colormap='Spectral')

mlab.view(90, 70, 6.2, (-1.3, -2.9, 0.25))
mlab.show()



import os
from mayavi import mlab
from tvtk.api import tvtk


def create_mb_dataset():
    mydir = os.path.dirname(__file__)
    path = os.path.join(mydir, os.pardir, 'data', 'fire_ug.vtu')
    r = tvtk.XMLUnstructuredGridReader(file_name=path)
    r.update()
    ug0 = r.output
    ug1 = tvtk.UnstructuredGrid()
    ug1.deep_copy(ug0)
    pts = ug1.points.to_array()
    pts[:, 2] += 3.0
    ug1.points = pts
    mb = tvtk.MultiBlockDataSet()
    mb.set_block(0, ug0)
    mb.set_block(1, ug1)
    return mb


def main():
    mb = create_mb_dataset()
    src = mlab.pipeline.add_dataset(mb)

    iso = mlab.pipeline.iso_surface(src)
    scp = mlab.pipeline.scalar_cut_plane(src)
    vcp = mlab.pipeline.vector_cut_plane(src)
    vcp.implicit_plane.widget.normal_to_y_axis = True

    mlab.show()


if __name__ == '__main__':
    main()



import unittest

try:
    from tvtk.pyface.ui.qt4.QVTKRenderWindowInteractor import (
        Key,
        QVTKRenderWindowInteractor,
        _repaint_after_render,
    )
except (ImportError, RuntimeError):
    # No binding installed (ImportError) or QT_API set but empty/invalid, as on
    # the headless CI row (RuntimeError) -- both mean "no Qt here".
    Key = QVTKRenderWindowInteractor = _repaint_after_render = None


@unittest.skipIf(_repaint_after_render is None, 'Qt is not available.')
class TestRepaintAfterRender(unittest.TestCase):
    """The macOS Qt >= 6.10 paint storm (see tvtk/WORKAROUNDS.md)."""

    f = staticmethod(_repaint_after_render)

    def test_darwin_qt_610_and_later_repaints_once(self):
        for version in ('6.10.0', '6.11.1', '7.0.0'):
            self.assertFalse(self.f('darwin', version), version)

    def test_darwin_older_qt_unaffected(self):
        for version in ('6.9.2', '5.15.10'):
            self.assertTrue(self.f('darwin', version), version)

    def test_other_platforms_unaffected(self):
        for platform in ('linux', 'win32'):
            self.assertTrue(self.f(platform, '6.11.1'), platform)

    def test_unparseable_version_unaffected(self):
        self.assertTrue(self.f('darwin', 'not.a.version'))


class _FakeAngleDelta:
    def __init__(self, y):
        self._y = y

    def y(self):
        return self._y


class _FakeWheelEvent:
    """Duck-typed QWheelEvent; Qt 5's also carries the deprecated delta()."""

    def __init__(self, y, with_delta=False):
        self._y = y
        if with_delta:
            self.delta = lambda: self._y

    def angleDelta(self):
        return _FakeAngleDelta(self._y)


@unittest.skipIf(QVTKRenderWindowInteractor is None, 'Qt is not available.')
class TestWheelEvent(unittest.TestCase):
    def test_wheel_events_reach_vtk(self):
        """Wheel events must reach VTK on Qt 5 and Qt 6 alike (gh-1276).

        Qt 5 wheel events still carry the deprecated delta() alongside
        angleDelta(), which used to divert them onto a legacy path whose
        coalescing timer was only ever created under Qt 4 -- so wheel
        zooming silently did nothing on PyQt5/PySide2.
        """
        from pyface.qt.QtGui import QApplication
        app = QApplication.instance() or QApplication([])  # noqa: F841
        w = QVTKRenderWindowInteractor()
        try:
            w._Iren.Enable()
            events = []
            w._Iren.AddObserver('MouseWheelForwardEvent',
                                lambda o, e: events.append('+'))
            w._Iren.AddObserver('MouseWheelBackwardEvent',
                                lambda o, e: events.append('-'))
            w.wheelEvent(_FakeWheelEvent(120, with_delta=True))  # Qt 5 notch
            w.wheelEvent(_FakeWheelEvent(-120, with_delta=True))
            w.wheelEvent(_FakeWheelEvent(30))   # Qt 6 trackpad partials...
            w.wheelEvent(_FakeWheelEvent(30))   # ...accumulate to a notch
            w.wheelEvent(_FakeWheelEvent(-20))  # never reaches a notch
            self.assertEqual(events, ['+', '-', '+'])
        finally:
            w.close()
            w.deleteLater()


class _FakeKeyEvent:
    def __init__(self, text, key=0):
        self._text = text
        self._key = key

    def text(self):
        return self._text

    def key(self):
        return self._key


@unittest.skipIf(QVTKRenderWindowInteractor is None, 'Qt is not available.')
class TestKeySyms(unittest.TestCase):
    def test_key_char_and_keysym(self):
        """Chars and non-printing keys must map to VTK keysyms."""
        # Use the module's Key alias: PyQt6 only has the scoped enum
        # (Qt.Key.Key_PageUp), flat Qt.Key_PageUp is an AttributeError there.
        from pyface.qt.QtGui import QApplication
        app = QApplication.instance() or QApplication([])  # noqa: F841
        w = QVTKRenderWindowInteractor()
        try:
            f = w._GetKeyCharAndKeySym
            self.assertEqual(f(_FakeKeyEvent('r')), ('r', 'r'))
            self.assertEqual(f(_FakeKeyEvent('+')), ('+', 'plus'))
            self.assertEqual(f(_FakeKeyEvent('?')), ('?', 'question'))
            # Non-printing keys have no text and map through the key code;
            # Prior/Next were dropped from the map for years.
            self.assertEqual(f(_FakeKeyEvent('', Key.Key_PageUp)),
                             ('\0', 'Prior'))
            self.assertEqual(f(_FakeKeyEvent('', Key.Key_PageDown)),
                             ('\0', 'Next'))
            self.assertEqual(f(_FakeKeyEvent('', Key.Key_Left)),
                             ('\0', 'Left'))
            self.assertEqual(f(_FakeKeyEvent('', Key.Key_Meta)),
                             ('\0', 'None'))
        finally:
            w.close()
            w.deleteLater()


if __name__ == '__main__':
    unittest.main()



"""
Script to render the images for the Mayavi user guide.

"""
# Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.


import os
import subprocess
import sys


#############################################################################
def use_ci_image_size():
    """ Renders at 1x, so images match the ones CI produces.

        savefig multiplies by the screen's device pixel ratio, and neither
        magnification nor an explicit size overrides it, so a HiDPI display
        yields 2x images.  QT_SCALE_FACTOR cancels it, but Qt reads it while
        building the QApplication -- hence the subprocess to learn the ratio
        before this process starts Qt.
    """
    if 'QT_SCALE_FACTOR' in os.environ:
        return
    probe = ('from pyface.qt import QtGui\n'
             'app = QtGui.QApplication([])\n'
             'print(app.primaryScreen().devicePixelRatio())\n')
    try:
        out = subprocess.run([sys.executable, '-c', probe], check=True,
                             capture_output=True, text=True).stdout
        ratio = float(out.strip().splitlines()[-1])
    except Exception as exc:
        print('Could not determine the device pixel ratio (%s); images may '
              'not match the ones CI renders' % exc)
        return
    if ratio != 1:
        print('Rendering at 1/%g to undo HiDPI scaling' % ratio)
        os.environ['QT_SCALE_FACTOR'] = str(1 / ratio)


# has to happen before the imports below, which build the QApplication; skipped
# when mlab_reference.py execs this file just to read IMAGE_DIR
if __name__ == '__main__':
    use_ci_image_size()

from mayavi import mlab                                            # noqa: E402
from mayavi.scripts import mayavi2                                 # noqa: E402

import numpy as np                                                 # noqa: E402
from inspect import getmembers                                     # noqa: E402
from PIL import Image, ImageChops                                  # noqa: E402

IMAGE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                    'mayavi', 'generated_images')



def trim(filename):
    """ Crops the uniform border off an image, like ``convert -trim``.

        Raises if what is left is too small to be a real figure: a scene that
        renders blank otherwise trims to a couple of pixels and ships as an
        illustration, which is how mlab.test_simple_surf went unnoticed for
        years.
    """
    with Image.open(filename) as img:
        img = img.convert('RGB')
        # the border colour is whatever the top-left pixel is, as ImageMagick
        # assumes with a fuzz of 0
        border = Image.new('RGB', img.size, img.getpixel((0, 0)))
        bbox = ImageChops.difference(img, border).getbbox()
        if bbox is None or min(bbox[2] - bbox[0], bbox[3] - bbox[1]) < 32:
            raise RuntimeError(
                '%s came out blank (drawn area %s in a %dx%d image); the scene '
                'probably renders nothing, or nothing the camera can see'
                % (filename, bbox, img.size[0], img.size[1]))
        if bbox != (0, 0) + img.size:
            img.crop(bbox).save(filename)


def capture_image(func, filename):
    """ Runs a function doing some mayavi drawing and save the resulting
        scene to a file.
    """
    mlab.clf()
    # several of the test functions plot random data, which would otherwise give
    # a different figure on every run
    np.random.seed(0)
    func()
    if not filename[-4:] in ('.jpg', '.png'):
        filename = '%s.jpg' % filename
    mlab.savefig(filename , size=(400, 400) )
    trim(filename)


def illustrate_module(module, directory=IMAGE_DIR):
    """ Uses the test functions of the modules (discovery ala nose) to
        create images for each function tested.
    """
    for name, func in getmembers(module):
        if not callable(func) or not name.lower().startswith('test'):
            continue
        if name.lower().endswith('anim'):
            continue
        # LaTeX doesn't like '.' in filename (sucks), so we replace them.
        filename = directory + os.sep + module.__name__.replace('.', '_') \
                        + '_' + name[5:]
        capture_image(func, filename=filename)

#############################################################################
# Entry point
@mlab.show
def main():
    from render_examples import keep_windows_in_background
    keep_windows_in_background()   # do not steal focus while rendering
    mlab.figure(bgcolor=(1, 1, 1), fgcolor=(0, 0, 0), size=(400, 400))
    print("Generating the mlab images...")
    illustrate_module(mlab)
    mayavi2.close()
    print("Done generating the mlab images")
    print("Generating the example pages")
    from render_examples import render_examples
    render_examples(render_images=True)
    print("Done generating the example pages")
    # main() is wrapped in mlab.show, which runs an event loop once it returns;
    # nothing closes the last window, so say explicitly that we are finished
    from pyface.qt import QtGui
    app = QtGui.QApplication.instance()
    if app is not None:
        app.quit()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
#
# Copyright (c) Enthought, Inc.
# All rights reserved.

from setuptools import Command, Distribution, setup
from setuptools.command.bdist_wheel import bdist_wheel
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop

import os
import sys
import time
import subprocess
import shutil
import re
from pathlib import Path

# NOTE: tvtk is imported lazily inside the build hooks below, not at module
# scope, so that metadata/sdist builds (which do not generate the TVTK ZIP)
# do not require tvtk to be importable.

MY_DIR = Path(__file__).resolve().parent

# The PEP 517 ``setuptools.build_meta`` backend does not put the source
# directory on sys.path (only the legacy ``:__legacy__`` variant does), so the
# in-tree ``tvtk`` package that the build hooks below import would not be
# importable during a build.  Put it on sys.path explicitly.
if str(MY_DIR) not in sys.path:
    sys.path.insert(0, str(MY_DIR))

DEFAULT_HTML_TARGET_DIR = Path('docs', 'build')
DEFAULT_INPUT_DIR = Path('docs', 'source')


class GenDocs(Command):

    description = (
        "This command generates generated part of the documentation "
        "when needed. It's run automatically before a build_docs, and that's "
        "the only time it needs to be run."
    )
    user_options = [
        ('None', None, 'this command has no options'),
        ]

    def latest_modified(self, the_path, filetypes='', ignore_dirs=''):
        """Traverse a path looking for the most recently modified file.

        Parameters
        ----------
        the_path : path-like
            Contains path to be traversed or filename to be inspected.
        filetypes : string
            Regular expression pattern of files to examine. If specified, other
            files are ignored. Otherwise, all files are examined.
        ignore_dirs : string
            Regular expression pattern of directories to be ignored. If ignore
            specified, all directories are walked.

        Returns
        -------
        latest_time : float
            Modification time of latest_path.
        latest_path : Path
            Most recently modified file.
        """
        the_path = Path(the_path)
        file_re = re.compile(filetypes)
        dir_re = re.compile(ignore_dirs)

        if not the_path.exists():
            return 0, the_path
        if the_path.is_dir():
            latest_time = 0
            latest_path = the_path
            # os.walk (rather than Path.rglob) so that ignored directories can
            # be pruned from the traversal in place
            for root, dirs, files in os.walk(the_path):
                root = Path(root)
                if ignore_dirs != '':
                    # This needs to iterate over a copy of the list. Otherwise,
                    # as things get removed from the original list, the indices
                    # become invalid.
                    for dir in dirs[:]:
                        if dir_re.search(dir):
                            dirs.remove(dir)
                for file in files:
                    if filetypes != '':
                        if not file_re.search(file):
                            continue
                    current_file_time = (root / file).stat().st_mtime
                    if current_file_time > latest_time:
                        latest_time = current_file_time
                        latest_path = root / file
            return latest_time, latest_path

        else:
            return the_path.stat().st_mtime, the_path

    def mlab_reference(self):
        """If mayavi is installed, run the mlab_reference generator."""
        # XXX: This is really a hack: the script is not made to be used
        # for different projects, but it ended up being. This part is
        # mayavi-specific.

        mlab_ref_dir = DEFAULT_INPUT_DIR / 'mayavi' / 'auto'

        source_path = Path('mayavi')
        sources = r'(\.py)|(\.rst)$'
        excluded_dirs = r'^\.'
        target_path = mlab_ref_dir
        target_time = self.latest_modified(target_path,
                                           ignore_dirs=excluded_dirs)[0]

        if (self.latest_modified(source_path, filetypes=sources,
                                 ignore_dirs=excluded_dirs)[0] > target_time
            or self.latest_modified('mlab_reference.py')[0] > target_time
                or not (mlab_ref_dir / 'mlab_reference.rst').exists()):
            try:
                from mayavi import mlab
                from mayavi.tools import auto_doc
                print("Generating the mlab reference documentation")
                os.system('python mlab_reference.py')
            except Exception:
                pass

    def example_files(self):
        """Generate the documentation files for the examples."""
        mlab_ref_dir = DEFAULT_INPUT_DIR / 'mayavi' / 'auto'

        source_path = Path('examples', 'mayavi')
        sources = r'(\.py)|(\.rst)$'
        excluded_dirs = r'^\.'
        target_path = mlab_ref_dir
        target_time = self.latest_modified(target_path,
                                           ignore_dirs=excluded_dirs)[0]

        script_file_name = DEFAULT_INPUT_DIR / 'render_examples.py'

        if (self.latest_modified(source_path, filetypes=sources,
                                 ignore_dirs=excluded_dirs)[0] > target_time
            or self.latest_modified(script_file_name)[0] > target_time
                or not (mlab_ref_dir / 'examples.rst').exists()):
            try:
                from mayavi import mlab
                from mayavi.tools import auto_doc
                print("Generating the example list")
                subprocess.call('python %s' % script_file_name.name,
                                shell=True, cwd=script_file_name.parent)
            except:
                pass

    def run(self):
        self.mlab_reference()
        self.example_files()

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class BuildDocs(Command):

    description = \
        "This command generates the documentation by running Sphinx. " \
        "It then zips the docs into an html.zip file."

    user_options = [
        ('None', None, 'this command has no options'),
        ]

    def make_docs(self):
        if os.name == 'nt':
            print("Please impelemnt sphinx building on windows here.")
        else:
            subprocess.call(['make', 'html'], cwd='docs')

    def run(self):
        self.make_docs()

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


def _tvtk_built_recently(zipfile, delay):
    """Returns True if the TVTK classes in zipfile was built in the last
    delay seconds.
    """
    if not zipfile.exists():
        return False

    ctime = zipfile.stat().st_ctime
    tdiff = time.time() - ctime
    return tdiff < delay


# Our custom distutils hooks
def build_tvtk_classes_zip():
    zipfile = MY_DIR / 'tvtk' / 'tvtk_classes.zip'
    if _tvtk_built_recently(zipfile, delay=120):
        print("Already built tvtk_classes.zip")
        return
    else:
        print("Building tvtk_classes.zip")
    from tvtk._setup import gen_tvtk_classes_zip
    gen_tvtk_classes_zip()


class MyBuildPy(build_py):
    """Build hook that generates ``tvtk_classes.zip`` and stages it into the
    build output.

    The ZIP is generated at build time from the installed VTK and is excluded
    from the sdist, so it cannot be collected as ordinary package data (it does
    not exist when the manifest is computed).  ``gen_tvtk_classes_zip`` writes
    it into the source tree next to the ``tvtk`` package, so copy it from there
    into ``build_lib`` explicitly; ``bdist_wheel`` then packages the whole tree.

    For editable installs the source tree *is* the install, so the ZIP written
    by ``build_tvtk_classes_zip`` is already in the right place and the copy
    below is a no-op as far as the installed package is concerned.
    """

    def run(self):
        build_tvtk_classes_zip()
        super().run()
        src = MY_DIR / 'tvtk' / 'tvtk_classes.zip'
        dst = Path(self.build_lib) / 'tvtk' / 'tvtk_classes.zip'
        self.mkpath(str(dst.parent))
        self.copy_file(str(src), str(dst))


class MyDevelop(develop):
    """A hook to build the TVTK ZIP file on develop."""

    def run(self):
        build_tvtk_classes_zip()
        super().run()


class MyDistribution(Distribution):
    """Claim ext modules so the whole build/install/wheel path uses the
    platlib layout: the code is pure Python, but the generated TVTK classes
    are platform-specific (X11 vs Cocoa vs Win32)."""

    def has_ext_modules(self):
        return True


class MyBdistWheel(bdist_wheel):
    """Tag wheels py3-none-<platform> (pure Python, platform-specific)."""

    def get_tag(self):
        _, _, plat = super().get_tag()
        # PyPI rejects plain linux_* tags; with no native code any glibc
        # floor is valid, so use manylinux_2_17 (understood by pip >= 20.3)
        if plat.startswith('linux_'):
            plat = plat.replace('linux_', 'manylinux_2_17_', 1)
        return 'py3', 'none', plat


###########################################################################
# Similar to package_data, but installed before build
build_package_data = {
    'mayavi.images': [DEFAULT_INPUT_DIR / 'mayavi' / '_static' / 'm2_about.jpg']
}

# Install our data files at build time. This is iffy,
# but we need to do this before distutils kicks in.
for package, files in build_package_data.items():
    target_path = Path(*package.split('.'))
    for filename in files:
        shutil.copy(filename, target_path)

###########################################################################

# Dependencies are dynamic (PEP 643): tvtk_classes.zip bakes in the
# build-time VTK API.  Older runtimes degrade gracefully but newer ones may
# remove API the generated code references, so wheels cap VTK at the next
# minor.  The sdist generates against the installer's VTK, so no cap.
DEPENDENCIES = [
    'apptools',
    'configobj',
    'envisage',
    'numpy',
    'pyface>=6.1.1',
    'pygments',  # only needed for the Qt backend but we add it anyway
    'traits>=6.0.0',
    'traitsui>=7.0.0',
    'packaging',
    "importlib_resources; python_version<'3.11'",
    'puremagic',
]


# Oldest supported VTK, per SPEC 0-style two-year support window.  When
# raising this, cull the workarounds it makes dead -- see tvtk/WORKAROUNDS.md.
MIN_VTK = '9.4'
# First VTK the generated wrappers are *not* known to work with.  Raise it
# once the vtk-dev CI row has been green against that version's prereleases.
MAX_VTK = '9.8'


def vtk_requirement():
    """The VTK requirement, with an upper bound when building a wheel."""
    if 'sdist' in sys.argv[1:]:
        return 'vtk>=%s' % MIN_VTK
    return 'vtk>=%s,<%s' % (MIN_VTK, MAX_VTK)


# Static metadata lives in pyproject.toml; setup.py carries only the dynamic
# dependencies and the imperative build hooks.
if __name__ == '__main__':
    setup(
        distclass=MyDistribution,
        cmdclass={
            'bdist_wheel': MyBdistWheel,
            'build_py': MyBuildPy,
            'develop': MyDevelop,
            'gen_docs': GenDocs,
            'build_docs': BuildDocs,
        },
        install_requires=DEPENDENCIES + [vtk_requirement()],
    )



"""Simple test for the Optional and Collection filters.
"""
# Author: Prabhu Ramachandran <prabhu [at] aero . iitb . ac . in>
# Copyright (c) Prabhu Ramachandran
# License: BSD Style.

# Standard library imports.
from os.path import abspath
from io import BytesIO
import copy

# Local imports.
from common import TestCase, get_example_data


class TestOptionalCollection(TestCase):

    def test(self):
        self.main()

    def do(self):
        ############################################################
        # Imports.
        script = self.script
        from mayavi.sources.vtk_file_reader import VTKFileReader
        from mayavi.filters.contour import Contour
        from mayavi.filters.optional import Optional
        from mayavi.filters.collection import Collection
        from mayavi.filters.api import PolyDataNormals
        from mayavi.modules.api import Surface

        ############################################################
        # Create a new scene and set up the visualization.
        s = self.new_scene()

        # Read a VTK (old style) data file.
        r = VTKFileReader()
        r.initialize(get_example_data('heart.vtk'))
        script.add_source(r)

        c = Contour()
        # `name` is used for the notebook tabs.
        n = PolyDataNormals(name='Normals')
        o = Optional(filter=n, label_text='Compute normals')
        coll = Collection(filters=[c, o], name='IsoSurface')
        script.add_filter(coll)
        s = Surface()
        script.add_module(s)

        ########################################
        # do the testing.
        def check(coll):
            """Check if test status is OK given the collection."""
            c, o = coll.filters
            c = c.filter
            n = o.filter
            assert coll.get_output_dataset().point_data.scalars.range == (127.5, 127.5)
            # Adding a contour should create the appropriate output in
            # the collection.
            c.contours.append(200)
            assert coll.get_output_dataset().point_data.scalars.range == (127.5, 200.0)
            # the collection's output should be that of the normals.
            assert coll.get_output_dataset() is n.get_output_dataset()
            # disable the optional filter and check.
            o.enabled = False
            assert 'disabled' in o.name
            assert coll.get_output_dataset() is c.get_output_dataset()
            # Set back everything to original state.
            c.contours.pop()
            o.enabled = True
            assert coll.get_output_dataset().point_data.scalars.range == (127.5, 127.5)
            assert coll.get_output_dataset() is n.get_output_dataset()
            assert 'disabled' not in o.name

        check(coll)

        ############################################################
        # Test if saving a visualization and restoring it works.

        # Save visualization.
        f = BytesIO()
        f.name = abspath('test.mv2') # We simulate a file.
        script.save_visualization(f)
        f.seek(0) # So we can read this saved data.

        # Remove existing scene.
        engine = script.engine
        engine.close_scene(s)

        # Load visualization
        script.load_visualization(f)
        s = engine.current_scene

        # Now do the check.
        coll = s.children[0].children[0]
        check(coll)

        ############################################################
        # Test if the Mayavi2 visualization can be deep-copied.

        # Pop the source object.
        source = s.children.pop()
        # Add it back to see if that works without error.
        s.children.append(source)
        # Now do the check.
        coll = s.children[0].children[0]
        check(coll)

        # Now deepcopy the source and replace the existing one with
        # the copy.  This basically simulates cutting/copying the
        # object from the UI via the right-click menu on the tree
        # view, and pasting the copy back.
        source1 = copy.deepcopy(source)
        s.children[0] = source1
        # Now do the check.
        coll = s.children[0].children[0]
        check(coll)

        # If we have come this far, we are golden!

if __name__ == "__main__":
    t = TestOptionalCollection()
    t.test()



"""Regenerate the docs' generated content (used by the Docs CI workflow).

Renders the gallery images and the example pages, then the mlab reference, in
that order: the generators only emit an ``.. image::`` for a figure that already
exists on disk, so running them the other way round silently drops the
illustrations.
"""

import faulthandler
import os
import re
import runpy
import sys
import warnings
from pathlib import Path

REPO = Path(__file__).resolve().parent.parent
SOURCE = REPO / 'docs' / 'source'

# (action, message prefix, category) -- the examples and the generators run as
# plain scripts, so this is the only thing making their warnings fatal; the
# Makefiles' -W covers Sphinx's own diagnostics, not Python's.
WARNING_FILTERS = (
    ('error', '', Warning),
    # third-party, reached through envisage's plugin manager and sphinxcontrib
    ('ignore', 'pkg_resources is deprecated as an API', UserWarning),
    ('ignore', 'Deprecated call to `pkg_resources.declare_namespace',
     DeprecationWarning),
    # unsatisfiable until pyface.workbench moves to apptools
    ('ignore', 'Workbench will be moved from pyface', PendingDeprecationWarning),
    # an example calling plt.show() is right; it is this renderer that has no
    # interactive matplotlib backend
    ('ignore', 'FigureCanvasAgg is non-interactive', UserWarning),
)


def apply_warning_filters():
    """Make warnings fatal, here and in the per-example child processes.

    `capture_in_subprocess` discards a child's output unless it exits
    non-zero, so a warning there is only ever seen by being raised.
    """
    filters = list(WARNING_FILTERS)
    # VTK's own numpy_support.vtk_to_numpy assigns to .shape, which NumPy 2.5
    # deprecated; fixed in 9.7, so keep it fatal there.  mayavi/tests/conftest.py
    # carries the same gate for the suites -- see tvtk/WORKAROUNDS.md
    from tvtk.common import vtk_major_version, vtk_minor_version
    if (vtk_major_version, vtk_minor_version) < (9, 7):
        filters.append(('ignore',
                        'Setting the shape on a NumPy array has been deprecated',
                        DeprecationWarning))
    for action, message, category in filters:
        warnings.filterwarnings(action, re.escape(message), category)
    # PYTHONWARNINGS matches the message as a literal prefix, not a regex
    os.environ['PYTHONWARNINGS'] = ','.join(
        ':'.join((action, message, category.__name__))
        for action, message, category in filters)


def main():
    # rendering drives VTK through a Qt event loop, where a crash arrives as a
    # bare signal number and nothing says where -- which is how qt_embedding's
    # SIGSEGV stayed opaque for as long as it did.  The variable carries this
    # into the per-example children.
    faulthandler.enable()
    os.environ['PYTHONFAULTHANDLER'] = '1'
    apply_warning_filters()

    # render_images.py imports its sibling render_examples, so docs/source has
    # to go on sys.path -- but docs/source/mayavi and docs/source/tvtk then
    # shadow the installed packages, as namespace packages win over an editable
    # install's finder no matter where on the path they turn up.  Binding the
    # real ones first settles it; neither import builds the QApplication, so
    # render_images.py can still set QT_SCALE_FACTOR when it runs.
    import mayavi  # noqa: F401
    import tvtk.api  # noqa: F401
    sys.path.insert(0, str(SOURCE))

    # both generators resolve their inputs relative to the working directory
    os.chdir(SOURCE)
    runpy.run_path(str(SOURCE / 'render_images.py'), run_name='__main__')

    os.chdir(REPO)
    runpy.run_path(str(REPO / 'mlab_reference.py'), run_name='__main__')

    # An example that fails to render keeps its committed image, so the run
    # looks clean and the figure quietly goes stale -- which is how lucy ended
    # up shipping an image CI had never produced.  The report is written rather
    # than raised so that the site and its images are still built and uploaded:
    # a red run is exactly when you want to look at them.
    import render_examples
    failures = render_examples.RENDER_FAILURES
    (REPO / 'render_failures.txt').write_text(
        ''.join(line + '\n' for line in failures))
    if failures:
        print('\n%d example(s) were meant to render and did not:'
              % len(failures), file=sys.stderr)
        for line in failures:
            print('  ' + line, file=sys.stderr)


if __name__ == '__main__':
    sys.exit(main())



"""Abstracts all VTK related modules into one module.  This makes it trivial to
support local VTK classes that a user may have built.

By default it imports all of VTK and then looks for a tvtk_local module and
imports everything from that.  In order to add local classes to the TVTK build
one may simply provide a tvtk_local.py module somewhere with any classes that
need to be wrapped.

"""

# Author: Prabhu Ramachandran <prabhu [at] aero.iitb.ac.in>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

from vtk import *
try:
    from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase
except ImportError:
    pass

try:
    from tvtk_local import *
except ImportError:
    pass


vtk_version = vtkVersion.GetVTKVersion()

# Remove classes that crash or hang when wrapped on a specific runtime VTK.
# Deleting the name here hides the class from code generation (code_gen.py
# checks hasattr) and from wrapping.  See tvtk/WORKAROUNDS.md for the full
# map of where VTK workarounds live.
if vtk_version in ["9.4.0", "9.4.1", "9.4.2"]:
    # Instantiating these using TVTK causes a crash on VTK 9.4.x so skipping.
    try:
        del vtkIOSSReader, vtkIOSSCellGridReader
    except NameError:
        pass
    if vtk_version == "9.4.2":
        # vtkXOpenGLRenderWindow segfaults when being deconstructed on 9.4.2
        try:
            del vtkXOpenGLRenderWindow
        except NameError:
            pass



import unittest

from tvtk.qt_x11 import embedding_error, x11_platform_override

WAYLAND_SESSION = {'WAYLAND_DISPLAY': 'wayland-0', 'DISPLAY': ':0'}


class TestX11PlatformOverride(unittest.TestCase):
    """Steering Qt onto XWayland (see tvtk/WORKAROUNDS.md)."""

    f = staticmethod(x11_platform_override)

    def test_wayland_session_steers_to_xcb(self):
        self.assertEqual(self.f(dict(WAYLAND_SESSION)), 'xcb')

    def test_explicit_qt_qpa_platform_wins(self):
        for choice in ('wayland', 'xcb', 'offscreen'):
            environ = dict(WAYLAND_SESSION, QT_QPA_PLATFORM=choice)
            self.assertIsNone(self.f(environ), choice)

    def test_x11_session_untouched(self):
        self.assertIsNone(self.f({'DISPLAY': ':0'}))

    def test_no_x_server_untouched(self):
        self.assertIsNone(self.f({'WAYLAND_DISPLAY': 'wayland-0'}))

    def test_windows_and_macos_sessions_untouched(self):
        # neither sets WAYLAND_DISPLAY, which is what the OS check would be
        self.assertIsNone(self.f({}))  # Windows
        self.assertIsNone(self.f({'DISPLAY': ':0'}))  # macOS with XQuartz

    def test_qt_toolkits_steered_others_untouched(self):
        for toolkit in ('qt4', 'qt', ''):
            environ = dict(WAYLAND_SESSION, ETS_TOOLKIT=toolkit)
            self.assertEqual(self.f(environ), 'xcb', toolkit)
        for toolkit in ('wx', 'null'):
            environ = dict(WAYLAND_SESSION, ETS_TOOLKIT=toolkit)
            self.assertIsNone(self.f(environ), toolkit)


class TestEmbeddingError(unittest.TestCase):
    """The error raised in place of the Xlib BadWindow abort."""

    f = staticmethod(embedding_error)

    def test_wayland_platforms_error(self):
        for name in ('wayland', 'wayland-egl'):
            error = self.f(name)
            self.assertIn('QT_QPA_PLATFORM=xcb', error)
            self.assertIn(name, error)

    def test_embeddable_platforms_pass(self):
        for name in ('xcb', 'offscreen', 'minimal', ''):
            self.assertIsNone(self.f(name), name)

    def test_other_oses_pass(self):
        self.assertIsNone(self.f('cocoa'))
        self.assertIsNone(self.f('windows'))


if __name__ == '__main__':
    unittest.main()



"""
Use the greedy-terrain-decimator to display a decimated terrain view.

This example illustrates decimating a terrain. We use the
greedy-terrain-decimator to create a reduced mesh with an optimized grid that
approximates the initial regular grid.

The initial grid is displayed in white, and the optimized grid is displayed in
black, with the surface it creates. The initial grid can be seen
disappearing as it goes under the surface of the approximated grid:
although the decimated mesh follows closely the original, it is not
exactly the same.

One can see that the reduction in number of polygons is huge: the white
grid is much finer than the black grid. It is interesting to note that
the decimated mesh follows closely the original mesh, including in number
of polygons, in spots where the terrain changes most quickly.

This example uses the Grand Canyon topological radar data, from NASA.

The greedy-terrain-decimator is only useful to decimate a surface
warped from 2D data. To decimated more general meshes, you can use the
less-efficient decimate-pro filter (see :ref:`example_julia_set_decimation`).
"""
# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) Enthought, Inc.
# License: BSD Style.

# Retrieve the grand Canyon topological data ###################################
# Original file:
#'ftp://e0srp01u.ecs.nasa.gov/srtm/version2/SRTM1/Region_04/N36W113.hgt.zip'
import os
if not os.path.exists('N36W113.hgt.zip'):
    # Download the data
    try:
        from urllib import urlopen
    except ImportError:
        from urllib.request import urlopen
    print('Downloading data, please wait (10M)')
    opener = urlopen(
    'https://s3.amazonaws.com/storage.enthought.com/www/sample_data/N36W113.hgt.zip'
        )
    open('N36W113.hgt.zip', 'wb').write(opener.read())

# Load the data (signed 2 byte integers, big endian) ###########################
import zipfile
import numpy as np

data = np.frombuffer(zipfile.ZipFile('N36W113.hgt.zip').read('N36W113.hgt'),
                    '>i2')
data = data.reshape((3601, 3601))
data = data[200:400, 1200:1400]
data = data.astype(np.float32)

# Plot an interecting section ##################################################
from mayavi import mlab
mlab.figure(1, size=(450, 390))
mlab.clf()
data = mlab.pipeline.array2d_source(data)

# Use a greedy_terrain_decimation to created a decimated mesh
terrain = mlab.pipeline.greedy_terrain_decimation(data)
terrain.filter.error_measure = 'number_of_triangles'
terrain.filter.number_of_triangles = 5000
terrain.filter.compute_normals = True

# Plot it black the lines of the mesh
lines = mlab.pipeline.surface(terrain, color=(0, 0, 0),
                                      representation='wireframe')
# The terrain decimator has done the warping. We control the warping
# scale via the actor's scale.
lines.actor.actor.scale = [1, 1, 0.2]

# Display the surface itself.
surf = mlab.pipeline.surface(terrain, colormap='gist_earth',
                                      vmin=1450, vmax=1650)
surf.actor.actor.scale = [1, 1, 0.2]

# Display the original regular grid. This time we have to use a
# warp_scalar filter.
warp = mlab.pipeline.warp_scalar(data, warp_scale=0.2)
grid = mlab.pipeline.surface(warp, color=(1, 1, 1),
                                      representation='wireframe')

mlab.view(-17, 46, 143, [1.46, 8.46, 269.4])

mlab.show()



# -*- coding: utf-8 -*-
#
# Mayavi documentation build configuration file, created by
# sphinx-quickstart on Sat Apr 12 23:25:24 2008.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# The contents of this file are pickled, so don't put values in the namespace
# that aren't pickleable (module imports are okay, they're removed automatically).
#
# All configuration values have a default value; values that are commented out
# serve to show the default value.


# Adding the current directory to the path, so that sphinx finds the
# extensions.
import faulthandler
import os
import re
import warnings
from pathlib import Path

# Python warnings during the build are fatal too: the Makefiles' -W only
# covers Sphinx's own diagnostics.  Example execution is separate -- see the
# filters in scripts/render_docs.py.
warnings.filterwarnings('error')

# autodoc imports mayavi, so VTK can take the build down with it; without this
# a crash is a bare signal number.  The variable carries it into any child.
faulthandler.enable()
os.environ['PYTHONFAULTHANDLER'] = '1'

try:
    # When translating the docs to another language, this variable is set 
    # elsewhere, so if it is not set, set it to a suitable default.
    basedir
except NameError:
    basedir = os.path.dirname(os.path.abspath(__file__))

# General configuration
# ---------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'traits.util.trait_documenter']

# mayavi.tools.server (autodoc'd by tips.rst) imports wx and twisted and calls
# wxreactor.install() at import time; neither belongs in a docs build
autodoc_mock_imports = ['wx', 'twisted']

# Base classes that trait_documenter emits but that nothing here documents:
# ETS packages we do not build docs for, and -- because the build runs with
# ETS_TOOLKIT=null -- the null toolkit's placeholder class.
nitpick_ignore = [
    ('py:class', 'mayavi.core.base.Base'),
    ('py:class', 'pyface.base_toolkit.Toolkit.__call__.<locals>.Unimplemented'),
    ('py:class', 'traits.has_traits.HasStrictTraits'),
    ('py:class', 'traits.has_traits.HasTraits'),
    ('py:class', 'tvtk.pyface.scene_model.SceneModel'),
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General substitutions.
project = 'mayavi'
copyright = 'Enthought, Inc.'

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
d = {}
fname = Path(basedir, '..', '..', '..', 'mayavi', '__init__.py')
exec(compile(fname.read_text(), str(fname), 'exec'), d)
# "4.8.4.dev", not "4.8.4.dev14+g8665ab72c.d20260729": the commit and date show
# up in the title of every page, so they would churn the whole site each build.
# A pre-release segment such as rc1 is kept.
version = release = re.sub(r'\.dev\d+', '.dev', d['__version__'].split('+')[0])

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# Files that are ``include``d into another document.  Sphinx would otherwise
# also read each of them as a document in its own right, which duplicates every
# label they define and evaluates their ``:func:`` roles without the
# ``currentmodule`` their host sets.
exclude_patterns = [
    'auto/changes.rst',         # -> changes.rst
    'examples.rst',             # -> application.rst
    'mlab_3d_plotting_functions.rst',
    'mlab_animating.rst',
    'mlab_case_studies.rst',
    'mlab_changing_object_looks.rst',
    'mlab_demo.rst',
    'mlab_figures_decorations.rst',
    'mlab_pipeline.rst',
    'mlab_running_scripts.rst',  # ^ all -> mlab.rst
    # Superseded by the hand-maintained mlab_pipeline_reference.rst, which is
    # what auto/mlab_reference.rst's toctree actually points at.
    'auto/mlab_pipeline_reference.rst',
]

# Each of those was published as a page of its own for years, so leave a
# meta-refresh stub at the old URL aimed at the document that now holds it.
moved_pages = {
    'auto/changes': 'changes',
    'auto/mlab_pipeline_reference': 'mlab_pipeline_reference',
    'examples': 'application',
    'mlab_3d_plotting_functions': 'mlab',
    'mlab_animating': 'mlab',
    'mlab_case_studies': 'mlab',
    'mlab_changing_object_looks': 'mlab',
    'mlab_demo': 'mlab',
    'mlab_figures_decorations': 'mlab',
    'mlab_pipeline': 'mlab',
    'mlab_running_scripts': 'mlab',
}

_STUB = """<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8" />
    <meta http-equiv="refresh" content="0; url=%(url)s" />
    <link rel="canonical" href="%(url)s" />
    <title>Page moved</title>
  </head>
  <body>
    <p>This page has moved to <a href="%(url)s">%(url)s</a>.</p>
  </body>
</html>
"""


def _write_moved_page_stubs(app, exception):
    import re
    from pathlib import Path, PurePosixPath

    if exception is not None or app.builder.name != 'html':
        return
    for old, new in moved_pages.items():
        old = PurePosixPath(old)
        # land on the section itself: docutils turns label underscores into
        # hyphens when it makes the id
        source = (Path(basedir) / old).with_suffix('.rst')
        # nothing after the colon, else it is a hyperlink target, not a label
        label = re.search(r'^\.\.\s+_([\w-]+):\s*$', source.read_text(), re.M)
        anchor = '#' + label.group(1).replace('_', '-') if label else ''
        url = '../' * (len(old.parts) - 1) + new + '.html' + anchor
        target = (Path(app.outdir) / old).with_suffix('.html')
        target.parent.mkdir(parents=True, exist_ok=True)
        target.write_text(_STUB % {'url': url})


def setup(app):
    app.connect('build-finished', _write_moved_page_stubs)
    # build-finished runs once in the main process, so -j is fine
    return {'parallel_read_safe': True, 'parallel_write_safe': True}

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# Options for HTML output
# -----------------------

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
html_theme = 'classic'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# shared with docs/site/index.html, which links it alongside classic.css
html_css_files = ['mayavi.css']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# off: with a build on every push this would restamp all ~115 pages daily.
# The site landing page carries the build date instead (see docs/Makefile).
html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Content template for the index page.
#html_index = ''

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = False

# If true, the reST sources are included in the HTML build as _sources/<name>.
#html_copy_source = True

# Output file base name for HTML help builder.
htmlhelp_basename = 'Mayavidoc'

# A logo displayed in the html sidebar.
html_logo = 'mayavi-logo.png'

# Options for LaTeX output
# ------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, document class [howto/manual]).
latex_documents = [('index', 'mayavi_user_guide.tex', 'Mayavi User Guide',
                        'Prabhu Ramachandran, Gael Varoquaux', 'manual')]

# Additional stuff for the LaTeX preamble.
latex_preamble = r"""
\definecolor{VerbatimColor}{rgb}{0.95,1,0.833}
\definecolor{VerbatimBorderColor}{rgb}{0.6,0.6,0.6}
"""

# A logo displayed on the cover page.
latex_logo = 'm2_about.jpg'


# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_use_modindex = True

# To specify names to enable gettext extracting and translation applying for i18n additionally. You can specify below names:
gettext_additional_targets = ['raw'] 

################################################################################
# A hack:
import shutil
shutil.copyfile(os.path.join(basedir, '../../CHANGES.txt'), os.path.join(basedir, './auto/changes.rst'))



"""
Using VTK to assemble a pipeline for segmenting MRI images. This example
shows how to insert well-controled custom VTK filters in Mayavi.

This example downloads an MRI scan, turns it into a 3D numpy array,
applies a segmentation procedure made of VTK filters to extract the
gray-matter/white-matter boundary.

The segmentation algorithm used here is very naive and should, of course,
not be used as an example of segmentation.

"""

### Download and unpack the data, if not already on disk ########################
import os
mri_slice_file = os.path.join('mri_data', 'MRbrain.1')
if not os.path.exists(mri_slice_file):
    if not os.path.exists('MRbrain.tar.gz'):
        # Download the data
        try:
            from urllib import urlopen
        except ImportError:
            from urllib.request import urlopen
        print("Downloading data, Please Wait (7.8MB)")
        opener = urlopen(
                    'http://graphics.stanford.edu/data/voldata/MRbrain.tar.gz')
        open('MRbrain.tar.gz', 'wb').write(opener.read())

    # Extract the data
    import tarfile
    tar_file = tarfile.open('MRbrain.tar.gz')
    try:
        os.mkdir('mri_data')
    except:
        pass
    tar_file.extractall('mri_data')
    tar_file.close()


### Read the data in a numpy 3D array ##########################################
import numpy as np
data = np.array([np.fromfile(os.path.join('mri_data', 'MRbrain.%i' % i),
                                        dtype='>u2') for i in range(1, 110)])
data = data.reshape((109, 256, 256))
data = data.T

################################################################################
# Heuristic for finding the threshold for the brain

# Exctract the percentile 20 and 80 (without using
# scipy.stats.scoreatpercentile)
sorted_data = np.sort(data.ravel())
l = len(sorted_data)
lower_thr = sorted_data[int(0.2*l)]
upper_thr = sorted_data[int(0.8*l)]

# The white matter boundary: find the densest part of the upper half
# of histogram, and take a value 10% higher, to cut _in_ the white matter
hist, bins = np.histogram(data[data > np.mean(data)], bins=50)
brain_thr_idx = np.argmax(hist)
brain_thr =  bins[brain_thr_idx + 4]

del hist, bins, brain_thr_idx

# Display the data #############################################################
from mayavi import mlab
from tvtk.api import tvtk

fig = mlab.figure(bgcolor=(0, 0, 0), size=(400, 500))
# to speed things up
fig.scene.disable_render = True

src = mlab.pipeline.scalar_field(data)
# Our data is not equally spaced in all directions:
src.spacing = [1, 1, 1.5]
src.update_image_data = True

#----------------------------------------------------------------------
# Brain extraction pipeline

# In the following, we create a Mayavi pipeline that strongly
# relies on VTK filters. For this, we make heavy use of the
# mlab.pipeline.user_defined function, to include VTK filters in
# the Mayavi pipeline.

# Apply image-based filters to clean up noise
thresh_filter = tvtk.ImageThreshold()
thresh_filter.threshold_between(lower_thr, upper_thr)
thresh = mlab.pipeline.user_defined(src, filter=thresh_filter)

median_filter = tvtk.ImageMedian3D()
try:
    median_filter.set_kernel_size(3, 3, 3)
except AttributeError:
    median_filter.kernel_size = [3, 3, 3]
median = mlab.pipeline.user_defined(thresh, filter=median_filter)

diffuse_filter = tvtk.ImageAnisotropicDiffusion3D(
                                    diffusion_factor=1.0,
                                    diffusion_threshold=100.0,
                                    number_of_iterations=5, )

diffuse = mlab.pipeline.user_defined(median, filter=diffuse_filter)

# Extract brain surface
contour = mlab.pipeline.contour(diffuse, )
contour.filter.contours = [brain_thr, ]

# Apply mesh filter to clean up the mesh (decimation and smoothing)
dec = mlab.pipeline.decimate_pro(contour)
dec.filter.feature_angle = 60.
dec.filter.target_reduction = 0.7

smooth_ = tvtk.SmoothPolyDataFilter(
                    number_of_iterations=10,
                    relaxation_factor=0.1,
                    feature_angle=60,
                    feature_edge_smoothing=False,
                    boundary_smoothing=False,
                    convergence=0.,
                )
smooth = mlab.pipeline.user_defined(dec, filter=smooth_)

# Get the largest connected region
connect_ = tvtk.PolyDataConnectivityFilter(extraction_mode=4)
connect = mlab.pipeline.user_defined(smooth, filter=connect_)

# Compute normals for shading the surface
compute_normals = mlab.pipeline.poly_data_normals(connect)
compute_normals.filter.feature_angle = 80

surf = mlab.pipeline.surface(compute_normals,
                                        color=(0.9, 0.72, 0.62))

#----------------------------------------------------------------------
# Display a cut plane of the raw data
ipw = mlab.pipeline.image_plane_widget(src, colormap='bone',
                plane_orientation='z_axes',
                slice_index=55)

mlab.view(-165, 32, 350, [143, 133, 73])
mlab.roll(180)

fig.scene.disable_render = False

#----------------------------------------------------------------------
# To make the link between the Mayavi pipeline and the much more
# complex VTK pipeline, we display both:
mlab.show_pipeline(rich_view=False)
from tvtk.pipeline.browser import PipelineBrowser
browser = PipelineBrowser(fig.scene)
browser.show()

mlab.show()



#!/usr/bin/env python
"""A simple example demonstrating the creation of actors and animating
the in a scene using visual modeule."""
# Author: Raashid Baig <raashid@aero.iitb.ac.in>
# License: BSD Style.

from math import sqrt

from tvtk.tools.visual import sphere, iterate, show, vector, curve

#Creating the actors for the scene
giant = sphere(pos=(-1.0e11, 0, 0),
               radius=2e10,
               color=(1, 0, 0),
               mass=2e30)

dwarf = sphere(pos=(1.5e11, 0, 0),
               radius=1e10,
               color=(1, 1, 0),
               mass=1e30)

giant.p = vector(0, 0, -1e4) * giant.mass
dwarf.p = -1*giant.p

# creating the curve which will trace the paths of actors
for a in [giant, dwarf]:
    a.orbit = curve(radius=2e9, color=a.color)

dt = 86400

def anim():
    #Creating the animation function which will be called at
    #uniform timeperiod through the iterate function
    dist = dwarf.pos - giant.pos
    force = 6.7e-11 * giant.mass * dwarf.mass * \
        dist/(sqrt(dist[0]**2 + dist[1]**2 + dist[2]**2))**3
    giant.p = giant.p + force*dt
    dwarf.p = dwarf.p - force*dt

    for a in [giant, dwarf]:
        a.pos = a.pos + (a.p/a.mass)*dt
        a.orbit.append(a.pos)
        p = a.orbit.points
        if len(p) > 1000:
            a.orbit.points = p[200:]

a = iterate(50, anim)
a.edit_traits()
show()
